from .point_to_area import *
from . import point_to_area
from . import mode
from .mode import *
from .saller import sal
from . import vgm
from . vgm import *
from .rigider import *
from . import rigider
from .fss import *
from . import fss
from .significance import *
from . import significance




import time
import sys
import copy
import numpy as np
import pandas as pd
from scipy import ndimage
from . import data_pre
from .centmatch import restructuring


def minsepfun(Id, dm0, dm1, indX, indXhat,nlat,nlon):
    #掩膜提取函数：data为原始数值数组，mask为获得掩码用的布尔数组f
    result_value = []
    Obs_value = []
    Fcst_value = []

    for i in range(len(Id[:, 0])):
        a = Id[i, 0]    #从id数组的第一列取值
        b = Id[i, 1]    #从id数组的第二列取值
       # print(a, b)
        Obsdata_mask = np.mat(dm0[a])    #需要掩膜的原始数据
        #print("labels_{}".format(b))
        Obs_mask = np.ones((nlat,nlon))
        Obs_mask[indXhat[b]] = 0
        #Obs_mask = indXhat["labels_{}".format(b)] < 1     #掩膜范围
        Obs_masked = np.ma.array(Obsdata_mask, mask = Obs_mask)    #返回值有三类masked_array，mask,fill_value
        Obs = np.min(Obs_masked)
        Obs_value.append(Obs)

        Fcstdata_mask = np.mat(dm1[b])
        Fcst_mask = np.ones((nlat,nlon))
        Fcst_mask[indX[a]] = 0
        #Fcst_mask = indX["labels_{}".format(a)] < 1
        Fcst = np.min(np.ma.array(Fcstdata_mask, mask = Fcst_mask))
        Fcst_value.append(Fcst)

        #print("a:" + str(a) + " b:"+str(b) + "  " + str(Obs)+" " + str(Fcst))

        result = min(Fcst, Obs)
        result_value.append(result)
    return result_value


def minboundmatch_bak(look_ff, mindist=float('inf'), show=False):
    mindist =10
    fun_type = "multiple"
    x = copy.deepcopy(look_ff)
    if show:
        begin_tiid = time.strftime("%Y-%m-%d %H:%M:%S", time.localtime(time.time()))
    if (type(x) != dict):  # R判断x是否为features,python 暂时判断x是否为dict
        try:
            sys.exit(0)
        except:
            print("minboundmatch: invalid x argument.")
    # if (x['Xlabelsfeature'] == None and x['Ylabelsfeature'] == None):
    if (x['grd_ob_features'] == None and x['grd_fo_features'] == None):
        try:
            sys.exit(0)
        except:
            print("minboundmatch: no features to match!")
    if (x['grd_ob_features'] == None):
        try:
            sys.exit(0)
        except:
            print("minboundmatch: no verification features to match.")
    if (x['grd_fo_features'] == None):
        try:
            sys.exit(0)
        except:
            print("minboundmatch: no model features to match.")
    fun_type = fun_type.lower()  # str.lower(),将type中的字母变为小写
    # type <- match.arg(type)    #未译
    out = x.copy()
    # a <- attributes(x)    #获取x中的属性
    out.update({'match_type': "minboundmatch"})
    match_message = {'match_message': "Matching based on minimum boundary separation using {} match".format(fun_type)}
    out.update({'match_message': match_message})
    # Xfeats = x['Xlabelsfeature']
    # Yfeats = x['Ylabelsfeature']
    Xfeats = x['grd_ob_features']
    Yfeats = x['grd_fo_features']
    Xfeats = data_pre.pick_labels(Xfeats)
    Yfeats = data_pre.pick_labels(Yfeats)
    if Xfeats != {}:
        n = len(Xfeats)
    else:
        n = 0
    if Yfeats != {}:
        m = len(Yfeats)
    else:
        m = 0
    if (m == 0 and n == 0):
        if show:
            print("\n", 'No features detected in either field.  Returning NULL.\n')
        return None
    elif (m == 0):
        if show:
            print("\n", 'No features detected in forecast field.  Returning NULL.\n')
        return None
    elif (n == 0):
        if show:
            print("\n", "No features detected in observed field.  Returning NULL.\n")
        return None
    rep00_1 = np.arange(1, n + 1, 1)
    rep00 = np.tile(rep00_1, m)
    rep0_1 = np.arange(1, m + 1, 1)
    rep0 = rep0_1.repeat(n, axis=0)  # 按行进行元素重复
    ind = np.vstack((rep00, rep0)).T  # 数组转置
    grid0 = look_ff["grid"]
    Xdmaps = {}
    Ydmaps = {}
    for i in range(1, n + 1):
        lable_index = np.zeros((grid0.nlat,grid0.nlon))
        lable_index[Xfeats[i]] = 1
        xdmaps = ndimage.morphology.distance_transform_edt(1 - lable_index)
        Xdmaps[i] = xdmaps
    for i in range(1, m + 1):
        lable_index = np.zeros((grid0.nlat,grid0.nlon))
        lable_index[Yfeats[i]] = 1
        ydmaps = ndimage.morphology.distance_transform_edt(1 - lable_index)
        Ydmaps[i] = ydmaps
    res = np.array(minsepfun(Id=ind, dm0=Xdmaps, dm1=Ydmaps, indX=Xfeats, indXhat=Yfeats,nlat=grid0.nlat,nlon=grid0.nlon))
    res = np.column_stack((ind, res))
    good = res[:, 2] <= mindist
    res = res[good]
    out.update({'values': pd.DataFrame(res, columns=["Observed Feature No.", "Forecast Feature No.",
                                                     "Minimum Boundary Separation"])})
    o = np.argsort(res[:, 2])
    res = res[o]
    # res = res_o.sort_values(by = "Minimum Boundary Separation").reset_index(drop = True)    #dataframe排序，默认升序，并更新index
    if (fun_type == "single"):
        N = len(res)
        Id = np.arange(N)
        Id = Id[o]
        # matches = res[res["Observed Feature No."] == res["Forecast Feature No."]].reset_index(drop = True)
        # matches = res_o[:len(Yfeats)]
        matches = np.array([])
        for i in range(N):
            matches0 = np.flipud(res[0, 0:2])  # 注意这个R里面取值以后，将预报放在前面，观测放在后面
            # res = res.reset_index(drop = True)     #res排序后，更新index，并且原来的index值不保存
            Id2 = (res[:, 0] == res[0, 0]) | (res[:, 1] == res[0, 1])
            res = res[~Id2]
            Id = Id[~Id2]
            matches = np.append(matches, matches0)
            if len(Id) == 0:
                break
        matches = matches.reshape(int(len(matches) / 2), 2)
    else:
        matches = np.fliplr(res[:, 0:2])
        matchlen = matches.shape[0]
        fuq = np.unique(matches[:, 0])
        flen = len(fuq)
        ouq = np.unique(matches[:, 1])
        olen = len(ouq)
        if (matchlen > 0):
            if (matchlen == flen and matchlen > olen):
                if (show):
                    print(
                        "Multiple observed features are matched to one or more forecast feature(s).  Determining implicit merges.\n")
            elif (matchlen > flen and matchlen == olen):
                if (show):
                    print(
                        "Multiple forecast features are matched to one or more observed feature(s).  Determining implicit merges.\n")
            elif (matchlen > flen and matchlen > olen):
                if (show):
                    print(
                        "Multiple matches have been found between features in each field.  Determining implicit merges.\n")
            elif (matchlen == flen and matchlen == olen):
                if (show):
                    print("No multiple matches were found.  Thus, no implicit merges need be considered.\n")
            out['implicit_merges'] = restructuring(matches[:, 0].astype(int) - 1,
                                                             matches[:, 1].astype(int) - 1)
        else:
            if (show):
                print("No objects matched.\n")
            out.update({"implicit_merges": None})

    # matches.insert(0, "Forecast Feature No.", matches.pop("Forecast Feature No."))    #先删除"Forecast Feature No."列，然后在原表中第0列插入被删掉的列,并重新命名
    matches = pd.DataFrame(matches, columns=(["Forecast", "Observed"]))
    matches = matches.sort_values(by=["Forecast"])  # R：将Forecast放在第一列，并以Forecast进行升序排列，python里面Forecast依然放在第二列
    out.update({"matches": matches})

    unmatched_X = set(ind[:, 0]) - set(matches["Observed"])  # 判断ind的值是否在matches里面，输出ind不在matches里面的值
    unmatched_Xhat = set(ind[:, 1]) - set(matches["Forecast"])
    unmatched = {"X": unmatched_X, "Xhat": unmatched_Xhat}
    out.update({"unmatched": unmatched})
    if (show):
        print(begin_tiid, '- begin.tiid')
    out.update({"MergeForced": False})
    out.update({"class": "matched"})

    return out


def minboundmatch(look_ff, mindist = float('inf'), show = False):
    fun_type = "single"
    #mindist =10
    #fun_type = "multiple" # 另一种匹配方式，设置该选项是，mindist需设置为一个有限值
    x = copy.deepcopy(look_ff)
    if show:
        begin_tiid = time.strftime("%Y-%m-%d %H:%M:%S", time.localtime(time.time()))
    if (type(x) != dict):    #R判断x是否为features,python 暂时判断x是否为dict
        try:
            sys.exit(0)
        except:
            print("minboundmatch: invalid x argument.")
    #if (x['Xlabelsfeature'] == None and x['Ylabelsfeature'] == None):
    if (x['grd_ob_features'] == None and x['grd_fo_features'] == None):
        try:
            sys.exit(0)
        except:
            print("minboundmatch: no features to match!")
    if (x['grd_ob_features'] == None):
        try:
            sys.exit(0)
        except:
            print("minboundmatch: no verification features to match.")        
    if (x['grd_fo_features'] == None):
        try:
            sys.exit(0)
        except:
            print("minboundmatch: no model features to match.")     
    fun_type = fun_type.lower()     # str.lower(),将type中的字母变为小写
    #type <- match.arg(type)    #未译
    out = x.copy()
    #a <- attributes(x)    #获取x中的属性
    out.update({'match_type':"minboundmatch"})
    #match_message = {'match_message' :"Matching based on minimum boundary separation using {} match".format(fun_type)}
    out.update({'match_message':"Matching based on minimum boundary separation using {} match".format(fun_type)})
    #Xfeats = x['Xlabelsfeature']
    #Yfeats = x['Ylabelsfeature']
    Xfeats = x['grd_ob_features']
    Yfeats = x['grd_fo_features']
    Xfeats = data_pre.pick_labels(Xfeats)
    Yfeats = data_pre.pick_labels(Yfeats)
    if Xfeats != {}:
        n = len(Xfeats)
    else:
        n = 0
    if Yfeats != {}:
        m = len(Yfeats)
    else:
        m = 0
    if (m == 0 and n == 0):
        if show:    
            print("\n",'No features detected in either field.  Returning NULL.\n')
        return None
    elif (m == 0):
        if show:
            print("\n",'No features detected in forecast field.  Returning NULL.\n')
        return None
    elif (n == 0):
        if show:
            print("\n", "No features detected in observed field.  Returning NULL.\n")
        return None
    rep00_1 = np.arange(1, n + 1, 1)
    rep00 = np.tile(rep00_1, m)
    rep0_1 = np.arange(1, m + 1, 1)
    rep0 = rep0_1.repeat(n, axis = 0)    #按行进行元素重复
    ind = np.vstack((rep00, rep0)).T    #数组转置
    
    Xdmaps = {}
    Ydmaps = {}
    grid0 = look_ff["grid"]
    for i in range(1, n + 1):
        lable_index = np.zeros((grid0.nlat,grid0.nlon))
        lable_index[Xfeats[i]] = 1
        xdmaps = ndimage.morphology.distance_transform_edt(1 - lable_index)
        Xdmaps[i] = xdmaps
    for i in range(1, m + 1):
        lable_index = np.zeros((grid0.nlat,grid0.nlon))
        lable_index[Yfeats[i]] = 1
        ydmaps = ndimage.morphology.distance_transform_edt(1 - lable_index)
        Ydmaps[i] = ydmaps

    #print(Xdmaps)
    res = np.array(minsepfun(Id = ind, dm0 = Xdmaps, dm1 = Ydmaps, indX = Xfeats, indXhat = Yfeats,nlat=grid0.nlat,nlon=grid0.nlon))
    #o = res.rank()

    res_o = np.column_stack((ind,res))
    res_o = pd.DataFrame(res_o, columns = ["ob", "fo", "Minimum Boundary Separation"])

    res_o["ob"] = res_o["ob"].astype(np.int16)
    res_o["fo"] = res_o["fo"].astype(np.int16)
    #print(res_o)
    #o = res.rank()
    #横轴是观测的目标序号，纵轴是预报的目标序号
    res = res.reshape(len(Yfeats), len(Xfeats))
    #以观测为准：观测→预报
    res_OF = np.argsort(res, axis = 0)
    A = np.arange(len(Xfeats))
    B = res_OF[0,:]    
    C = np.array((A, B)).T
    res_OF = pd.DataFrame(C, columns = ['ob', 'fo'])
    #以预报为准：预报→观测,R里面应该是以预报为准
    res_FO = np.argsort(res, axis = 1)
    A = np.arange(len(Yfeats))
    B = res_FO[:,0]    
    C = np.array((A,B)).T+1
    res_FO = pd.DataFrame(C, columns = ['ob', 'fo'])

    
    #good = res["Minimum Boundary Separation"] <= mindist    #判断计算结果是否小于设置的参数mindist
    #res <- res[good, , drop = False]    #如果res中存在异常值，删除
    res_o = res_o[res_o["Minimum Boundary Separation"] <= mindist].reset_index(drop = True)
    out.update({'values': res_o})
    
    #获取所有匹配中，值最小的前几个(不以观测或者预报为准)
    o = res_o["Minimum Boundary Separation"].rank()    #dataframe排序，不改变数值的位置，标记排名
    res_o = res_o.sort_values(by = "Minimum Boundary Separation").reset_index(drop = True)    #dataframe排序，默认升序，并更新index

    if (fun_type == "single"):
        N = len(res)
        Id = np.arange(N)
        Id = o
        #matches = res[res["ob"] == res["fo"]].reset_index(drop = True)
        matches = res_o[:len(Yfeats)]
        '''
        #逻辑走得通，但是没循环到
        matches = np.zeros((1,2))
        for i in N:
            matches = res[:1]    #选取res的第一行，也是值最小的配对
            res = res.reset_index(drop = True)     #res排序后，更新index，并且原来的index值不保存
            Id2 = res[(res["ob"] == res["ob"][0]) | \
                    (res["fo"] == res["fo"][0])]
            Id = Id2[Id2 != True].index    #Id2[Id2 != True].index.tolist #获取index值
            if len(Id) == 0 :
                break
        '''
    else:

        matches = res_o[["ob","fo"]]    #获取dataframe前两列
        matchlen = matches.shape[0]
        fuq =  set((matches["fo"]))   #预报场set存放,使得元素不重复
        flen = len(fuq)
        ouq = set((matches["ob"]))    #观测场set存放,使得元素不重复
        olen = len(ouq)
        if (matchlen > 0):
            if (matchlen == flen and matchlen > olen):
                if (show):
                    print("Multiple observed features are matched to one or more forecast feature(s).  Determining implicit merges.\n")
            elif (matchlen > flen and matchlen == olen):
                if (show):
                    print("Multiple forecast features are matched to one or more observed feature(s).  Determining implicit merges.\n")
            elif (matchlen > flen and matchlen > olen):
                if (show):
                    print("Multiple matches have been found between features in each field.  Determining implicit merges.\n")
            elif (matchlen == flen and matchlen == olen):
                if (show):
                    print("No multiple matches were found.  Thus, no implicit merges need be considered.\n")
        else :
            if (show):
                print("No objects matched.\n")
            out.update({"implicit_merges": None})

    #matches.insert(0, "fo", matches.pop("fo"))    #先删除"fo"列，然后在原表中第0列插入被删掉的列,并重新命名
    #matches[["ob", "fo"]] = matches[["fo", "ob"]]    #R:res[, 2:1]表示先去第二列，再取第一列，所以重新命名，python此处调换两列顺序
    #matches.columns = ["fo", "ob", "Minimum Boundary Separation"]    #列名调换
    matches = matches.sort_values(by = ["fo"])    #R：将Forecast放在第一列，并以Forecast进行升序排列，python里面Forecast依然放在第二列


    unmatched_X = set(ind[:, 0]) - set(matches["ob"])    #判断ind的值是否在matches里面，输出ind不在matches里面的值
    unmatched_Xhat = set(ind[:, 1]) - set(matches["fo"])
    unmatched = {"ob":unmatched_X, "fo":unmatched_Xhat}
    out.update({"unmatched": unmatched})
    if (show):
        print(begin_tiid,'- begin.tiid')
    out.update({"MergeForced": False})
    out.update({"class":"matched"})
    #print(matches)
    out.update({"matches":matches[["fo","ob"]].values})
    merges = restructuring(matches['fo'], matches['ob'])
    out.update({"merges":merges})

    return out


#=============================  Example  ===================================
'''
hold = make_SpatialVx_PA1.makeSpatialVx(X = pd.read_csv("F:\\Work\\MODE\\tra_test\\FeatureFinder\\pert000.csv"), \
                    Xhat = pd.read_csv(r"F:\Work\MODE\tra_test\FeatureFinder\pert004.csv"), \
                    loc = pd.read_csv(r"F:\Work\MODE\tra_test\FeatureFinder\ICPg240Locs.csv"), \
                    thresholds = [0.01, 20.01], projection = True, subset = None, timevals = None, reggrid = True,\
                    Map = True, locbyrow = True, fieldtype = "Precipitation", units = ("mm/h"), dataname = "ICP Perturbed Cases", obsname = "pert000", \
                    modelname = "pert004" , q = (0, 0.1, 0.25, 0.33, 0.5, 0.66, 0.75, 0.9, 0.95), qs = None)

look_FeatureFinder = FeatureFinder_test_PA3.featureFinder(Object = hold, smoothfun = "disk2dsmooth", \
                     dosmooth = True, smoothpar = 17, smoothfunargs = None,\
                     thresh = 310, idfun = "disjointer", minsize = np.array([1]),\
                     maxsize = float("Inf"), fac = 1, zerodown = False, timepoint = 1,\
                     obs = 1, model = 1)
'''
'''
x = look.copy()    #FeatureFinder函数结果
#fun_type = str(["single", "multiple"])     
fun_type = "single"    #类型只能为"single"或者 "multiple"，如果是single,matches结果为一对一（长度为5），multiple的结果为一对多（长度为5*5）
mindist = 20
show = False

look_minboundmatch = minboundmatch(x = look_FeatureFinder.copy(), fun_type = "single", mindist = 20, show = False)

'''




from .time_compare import *
from . import data_prepare
from . fun import get_hour_dhour_list,get_dati_of_path,get_dati_str_of_path,get_ob_time_range_str,get_time_range_str
from .data_collection import download_from_gds,remove
#from .read_data_base_on_gds import read_stadata,read_griddata,read_stadata_from_griddata,read_stawind,read_stawind_from_gridwind
from .data_prepare import creat_fo_dataset,creat_ob_dataset,prepare_dataset,prepare_dataset_without_combining
from .terrain_height_correction import terrain_height_correct


import datetime
import numpy as np
import math
from scipy import optimize as optimize
from . import imomenter, fint2d, rigid_transform, q_loss_rigid
import meteva


def rigid_optimal(grd_ob,grd_fo, translate=True, rotate=True,stages=True, show=False):
    x = grd_fo.values.squeeze()
    y = grd_ob.values.squeeze()
    result = rigider(x,y,init = None,translate = translate,rotate = rotate,
                         stages = stages,show = show)

    grid0 = meteva.base.get_grid_of_data(grd_fo)
    rire = {}
    rire["grd_fo"] = grd_fo
    rire["grd_ob"] = grd_ob
    rire["grd_fo_transformed"] = meteva.base.grid_data(grid0,result["grd_fo_transformed"])
    meteva.base.set_griddata_coords(rire["grd_fo_transformed"],member_list=["变换后"])
    if "y" in result["par"].keys():
        rire["delta_lon"] = result["par"]["y"] * grid0.dlon
    else:
        rire["delta_lon"] = 0
    if "x" in result["par"].keys():
        rire["delta_lat"] = result["par"]["x"] * grid0.dlat
    else:
        rire["delta_lat"] = 0
    if "theta" in result["par"].keys():
        rire["delta_angle"] = result["par"]["theta"] * 180 / 3.1415926
    else:
        rire["delta_angle"] = 0
    rire["rmse_before"] = np.sqrt(np.mean(np.power(grd_fo.values - grd_ob.values,2)))
    rire["rmse_after"] = np.sqrt(np.mean(np.power(rire["grd_fo_transformed"].values - grd_ob.values, 2)))

    return rire


def rigid_simple(grd_ob,grd_fo):
    x = grd_fo.values.squeeze()
    y = grd_ob.values.squeeze()
    result = rigider(x,y,func_type="fast",stages=False)

    grid0 = meteva.base.get_grid_of_data(grd_fo)
    rire = {}
    rire["grd_fo"] = grd_fo
    rire["grd_ob"] = grd_ob
    rire["grd_fo_transformed"] = meteva.base.grid_data(grid0,result["grd_fo_transformed"])
    meteva.base.set_griddata_coords(rire["grd_fo_transformed"],member_list=["变换后"])
    if "y" in result["par"].keys():
        rire["delta_lon"] = result["par"]["y"] * grid0.dlon
    else:
        rire["delta_lon"] = 0
    if "x" in result["par"].keys():
        rire["delta_lat"] = result["par"]["x"] * grid0.dlat
    else:
        rire["delta_lat"] = 0
    if "theta" in result["par"].keys():
        rire["delta_angle"] = result["par"]["theta"] * 180 / 3.1415926
    else:
        rire["delta_angle"] = 0
    rire["rmse_before"] = np.sqrt(np.mean(np.power(grd_fo.values - grd_ob.values,2)))
    rire["rmse_after"] = np.sqrt(np.mean(np.power(rire["grd_fo_transformed"].values - grd_ob.values, 2)))

    return rire


def plot_value(rire,cmap = "rain_1h",clevs = None):
    grd_list = [rire["grd_ob"],rire["grd_fo"],rire["grd_fo_transformed"]]
    vmax = max(np.max(grd_list[0].values),np.max(grd_list[0].values))
    vmin = min(np.min(grd_list[0].values),np.min(grd_list[0].values))
    cmap1, clevs1 = meteva.base.tool.color_tools.def_cmap_clevs(cmap=cmap, clevs=clevs, vmin=vmin, vmax=vmax)
    meteva.base.plot_tools.plot_2d_grid_list(grd_list,cmap=cmap1,clevs = clevs1,ncol= 3)
    return

def rigider(grd_fo, grd_ob, init=None, func_type="regular", translate=True, rotate=True,
            interp="linear", stages=True, show=False):
    if stages and not (translate and rotate):
        stages = False
    if not translate and not rotate:
        raise Exception("rigider: one of translate or rotate must be true.")
    if show:
        begin_tiid = datetime.datetime.now()
    x_dim = grd_ob.shape[0]
    y_dim = grd_ob.shape[1]
    #range0 = np.tile(np.arange(1, x_dim + 1), y_dim)
    #range1 = (np.arange(1, y_dim + 1)).repeat(x_dim)
    range0 = np.tile(np.arange(0, x_dim), y_dim)
    range1 = (np.arange(0, y_dim)).repeat(x_dim)
    p0 = np.stack((range0, range1), axis=-1)
    out = {}
    if isinstance(func_type, str):
        func_type = func_type.lower()
    elif isinstance(func_type, list):
        func_type = func_type[0].lower()
    p0 = np.array(p0)
    big_n = p0.shape[0]
    grd_fo = np.array(grd_fo)
    # 计算质心
    array_index = np.array(grd_fo.transpose() > 0).reshape((grd_fo.size, 1)).repeat(2, axis=1)
    temp_matrix = p0[array_index]
    temp_array = np.mean(temp_matrix.reshape((int(temp_matrix.size / 2), 2)), axis=0)
    field_center = np.tile(np.array(temp_array), big_n).reshape(big_n, 2, order='C')
    if func_type == "regular":
        xdim = grd_fo.shape
        if xdim != grd_ob.shape:
            raise Exception("rigider: x1 and x0 must have the same dimension.")
        outpar = np.zeros(1) + np.nan
        if init is None:
            if show:
                print("Initial values not passed.  Determining good starting values now.\n")
            # 计算图片运动参数（images movements）, 获取刚体变换的初始平移量和旋转量
            # reference: Hu, M. K. (1962) Visual Pattern Recognition by Moment Invariants. IRE Trans. Info. Theory, IT-8, 179–187.
            hold1 = imomenter.imomenter(grd_fo, loc=p0)
            hold0 = imomenter.imomenter(grd_ob, loc=p0)
            init = [hold1["centroid"]["x"] - hold0["centroid"]["x"], hold1["centroid"]["y"] - hold0["centroid"]["y"],
                  hold1["orientation_angle"] - hold0["orientation_angle"]]
            if not translate:
                init[0] =0
                init[1] =0
            if not rotate:
                init[2] = 0
            if show:
                print("initial values:\n")
                print(init)
        if not stages:
            if translate:
                if show:
                    print("Optimizing translation.\n")
                # 使用优化函数获取最佳变量
                res = optimize.minimize(ofun1, init[0:2],
                                        (p0, grd_fo, grd_ob, interp, big_n, field_center, xdim),"L-BFGS-B")
                if show:
                    print("Optimal translation found to be: ", res["x"], "\nwhere loss value is: ", res["fun"],
                          "\n")
                temp_theta = np.append(res["x"], init[2])
                p1 = rigid_transform.rigid_transform(theta=temp_theta, p0=p0, n=big_n, cen=field_center)
                y1 = fint2d.fint2d(x=grd_fo, ws=p1, s=p0, method=interp)
                res["p1"] = p1
                out["grd_fo_translated"] = y1
                outpar = res["x"]
                outval = res["fun"]
            if rotate:
                if translate:
                    init2 = np.append(res["x"], init[2])
                else:
                    init2 = init
                if show:
                    print("Optimizing rotation.\n")
                res2 = optimize.minimize(ofun2, init2[2],
                                         (p0, grd_fo, grd_ob, interp, big_n, field_center, xdim,
                                          init2[0:2]), "L-BFGS-B")
                if show:
                    print("Optimal rotation found to be: ", res2["x"], "\nwhere loss value is: ", res2["fun"], "\n")

                if translate:
                    r_center = np.tile(np.array([hold0["centroid"]["x"], hold0["centroid"]["y"]]), big_n).reshape(big_n, 2,
                                                                                                                  order='C')
                else:
                    r_center = np.tile(np.array([hold1["centroid"]["x"], hold1["centroid"]["y"]]), big_n).reshape(big_n, 2,
                                                                                                              order='C')
                p1 = rigid_transform.rigid_transform(theta=np.array([init2[0],init2[1], res2["x"]]), p0=p0, n=big_n,
                                                     cen=r_center)


                y1 = fint2d.fint2d(x=grd_fo, ws=p1, s=p0, method=interp)
                outpar = np.append(outpar, res2["x"])
                outval = res2["fun"]
        else:
            if show:
                print("Optimizing rigid transformation.\n")
            res = optimize.minimize(ofun3, init, (p0, grd_fo, grd_ob, interp, big_n, field_center, xdim),
                                    "L-BFGS-B")
            if show:
                print("Optimal transformation found to be: ", res["x"], "\nwhere loss value is: ", res["fun"], "\n")
            r_center = np.tile(np.array([hold0["centroid"]["x"], hold0["centroid"]["y"]]), big_n).reshape(big_n, 2,
                                                                                                              order='C')
            p1 = rigid_transform.rigid_transform(theta=res["x"], p0=p0, n=big_n, cen=r_center)

            outpar = res["x"]
            outval = res["fun"]

        if translate and (not rotate):
            outpar = {"x": outpar[0], "y": outpar[1]}
        elif (not translate) and rotate:
            outpar = {"angle": outpar[1]} # 注意 前面outpar的设置
        else:
            outpar = {"x": outpar[0], "y": outpar[1], "theta": outpar[2]}
        if not stages:
            if translate:
                out["translation_only"] = res
            if rotate:
                out["rotate"] = res2
        else:
            out["optim_object"] = res
    elif func_type == "fast":
        hold1 = imomenter.imomenter(grd_fo, loc=p0)
        hold0 = imomenter.imomenter(grd_ob, loc=p0)
        #print(hold0)
        #print(hold1)
        if translate:
            tr = [hold1["centroid"]["x"] - hold0["centroid"]["x"],hold1["centroid"]["y"] - hold0["centroid"]["y"]]
        #print(tr)
        if rotate:
            rot = hold1["orientation_angle"] - hold0["orientation_angle"]
        if translate and rotate:
            outpar = np.array([tr[0],tr[1], rot])
            inpar = np.array([tr[0],tr[1], rot])
            outpar = {"x": outpar[0], "y": outpar[1], "theta": outpar[2]}
        elif (not translate) and rotate:
            outpar = rot
            inpar = np.array([0, 0, rot])
            outpar = {"theta": outpar[0]}
        elif translate and (not rotate):
            outpar = tr
            outpar = {"x": outpar[0], "y": outpar[1]}
            inpar = np.array([tr, 0])
        if stages:
            p1_tr = rigid_transform.rigid_transform(theta=tr, p0=p0, n=big_n, cen=field_center)
            y1_tr = fint2d.fint2d(x=grd_fo, ws=p1_tr, s=p0, method=interp)
            out["grd_fo_translated"] = y1_tr
        # 计算刚体变换后的坐标
        #print(field_center)
        ob_center = np.tile(np.array([hold0["centroid"]["x"],hold0["centroid"]["y"]]), big_n).reshape(big_n, 2, order='C')
        p1 = rigid_transform.rigid_transform(theta=inpar, p0=p0, n=big_n, cen=ob_center)
        #p1 = rigid_transform.rigid_transform(theta=inpar, p0=p0, n=big_n, cen=field_center)

    # 计算刚体变换后的x1
    y1 = fint2d.fint2d(x=grd_fo, ws=p1, s=p0, method=interp)
    #y1 = fint2d.fint2d(x=grd_fo, ws=p0, s=p1, method=interp)
    out["func_type"] = func_type
    if func_type == "regular":
        out["initial"] = init
        out["value"] = outval
    out["interp"] = interp
    out["par"] = outpar
    out["grd_ob"] = grd_ob
    out["grd_fo"] = grd_fo
    out["p0"] = p0
    out["p1"] = p1
    out["grd_fo_transformed"] = y1
    if show:
        print(datetime.datetime.now() - begin_tiid)
    return out


def ofun1(theta, p0, x1, x0, interp, n, cen, xdim, tr=None):
    if theta[0] > xdim[0] / 2 + 1 or theta[1] > xdim[1] / 2 + 1:
        return 1e+16
    theta = np.array(theta)
    theta = np.append(theta, 0)
    p1 = rigid_transform.rigid_transform(theta=theta, p0=p0, n=n, cen=cen)
    y1 = fint2d.fint2d(x=x1, ws=p1, s=p0, method=interp)
    res = q_loss_rigid.q_loss_rigid(y1, x0, p1, p0)
    return res


def ofun2(theta, p0, x1, x0, interp, n, cen, xdim, tr):
    if theta > math.pi / 2 or theta < -math.pi / 2:
        return 1e+16
    theta = np.array(theta)
    theta = np.append(tr, theta)
    p1 = rigid_transform.rigid_transform(theta=theta, p0=p0, n=n, cen=cen)
    y1 = fint2d.fint2d(x=x1, ws=p1, s=p0, method=interp)
    res = q_loss_rigid.q_loss_rigid(y1, x0, p1, p0)
    return res


def ofun3(theta, p0, x1, x0, interp, n, cen, xdim, tr=None):
    if theta[0] > xdim[0] / 2 + 1 or theta[1] > xdim[1] / 2 + 1:
        return 1e+16
    condition_a = theta > math.pi / 2
    condition_b = theta < -math.pi / 2
    condition = condition_a + condition_b
    if condition.all():
        #if theta > math.pi / 2 or theta < -math.pi / 2:
        return 1e+16
    p1 = rigid_transform.rigid_transform(theta=theta, p0=p0, n=n, cen=cen)
    y1 = fint2d.fint2d(x=x1, ws=p1, s=p0, method=interp)
    res = q_loss_rigid.q_loss_rigid(y1, x0, p1, p0)
    return res


if __name__ == '__main__':
    import meteva.base as meb
    import meteva.method as mem
    # 生成方形的测试数据
    grid_ob = meb.grid([100, 130, 1], [15, 45, 1], gtime=["2021080408"], dtime_list=[0], member_list=["OBS"])
    x = np.zeros((31, 31))
    x[8:10, 10:17] = 10
    grd_ob = meb.grid_data(grid_ob, x)

    grid_fo = meb.grid([100, 130, 1], [15, 45, 1], gtime=["2021080308"], dtime_list=[24], member_list=["ECMWF"])
    y = np.zeros((31, 31))
    for i in range(y.shape[0]):
        for j in range(y.shape[1]):
            a = (i - 15) * math.sin(math.pi / 4) - (j - 15) * math.cos(math.pi / 4)
            b = (i - 15) * math.sin(math.pi / 4) + (j - 15) * math.cos(math.pi / 4)
            if abs(a) < 3 and abs(b) < 5:
                y[i, j] = 5
    grd_fo = meb.grid_data(grid_fo, y)
    rire = mem.rigider.rigid_simple(grd_ob, grd_fo)
    mem.rigider.plot_value(rire)


from . import short_term_heavy_rainfall
from . import thunderstrom_gale
from . import temperature

from .short_term_heavy_rainfall import *
from .thunderstrom_gale import *
from .temperature import  *


import numpy as np
import matplotlib.image as image
from matplotlib import cm
import matplotlib.pyplot as plt
plt.rcParams['font.sans-serif']=['SimHei'] #用来正常显示中文标签
plt.rcParams['axes.unicode_minus']=False #用来正常显示负号
import matplotlib.colors as colors
import pkg_resources
import math
import os
import colorsys
import meteva

def clev_cmap_temper_2m_k():
    path = pkg_resources.resource_filename('meteva', "resources/colormaps/color_temp_2m.txt")
    cmap,clevs =  get_cmap_and_clevs_from_file(path)
    clevs += 273.15
    print("不再推荐使用该函数，推荐使用meb.def_cmap_clevs相应功能,请参考color工具中相关说明")
    return clevs,cmap

def clev_cmap_temper_2m():
    path = pkg_resources.resource_filename('meteva', "resources/colormaps/color_temp_2m.txt")
    cmap, clevs = get_cmap_and_clevs_from_file(path)
    print("不再推荐使用该函数，推荐使用meb.def_cmap_clevs相应功能,请参考color工具中相关说明")
    return clevs, cmap

def clev_cmap_rain_1h():
    path = pkg_resources.resource_filename('meteva', "resources/colormaps/color_rain_1h.txt")
    cmap, clevs = get_cmap_and_clevs_from_file(path)
    print("不再推荐使用该函数，推荐使用meb.def_cmap_clevs相应功能,请参考color工具中相关说明")
    return clevs, cmap
def clev_cmap_rain_3h():
    path = pkg_resources.resource_filename('meteva', "resources/colormaps/color_rain_3h.txt")
    cmap, clevs = get_cmap_and_clevs_from_file(path)
    print("不再推荐使用该函数，推荐使用meb.def_cmap_clevs相应功能,请参考color工具中相关说明")
    return clevs, cmap
def clev_cmap_rain_24h():
    path = pkg_resources.resource_filename('meteva', "resources/colormaps/color_rain_24h.txt")
    cmap, clevs = get_cmap_and_clevs_from_file(path)
    print("不再推荐使用该函数，推荐使用meb.def_cmap_clevs相应功能,请参考color工具中相关说明")
    return clevs, cmap
def clev_cmap_rh():
    path = pkg_resources.resource_filename('meteva', "resources/colormaps/color_rh.txt")
    cmap, clevs = get_cmap_and_clevs_from_file(path)
    print("不再推荐使用该函数，推荐使用meb.def_cmap_clevs相应功能,请参考color工具中相关说明")
    return clevs, cmap
def clev_cmap_vis():
    path = pkg_resources.resource_filename('meteva', "resources/colormaps/color_vis.txt")
    cmap, clevs = get_cmap_and_clevs_from_file(path)
    print("不再推荐使用该函数，推荐使用meb.def_cmap_clevs相应功能,请参考color工具中相关说明")
    return clevs, cmap
def clev_cmap_wind_speed():
    path = pkg_resources.resource_filename('meteva', "resources/colormaps/color_wind_speed.txt")
    cmap, clevs = get_cmap_and_clevs_from_file(path)
    print("不再推荐使用该函数，推荐使用meb.def_cmap_clevs相应功能,请参考color工具中相关说明")
    return clevs, cmap
def clev_cmap_cloud_total():
    path = pkg_resources.resource_filename('meteva', "resources/colormaps/color_tcdc.txt")
    cmap, clevs = get_cmap_and_clevs_from_file(path)
    print("不再推荐使用该函数，推荐使用meb.def_cmap_clevs相应功能,请参考color工具中相关说明")
    return clevs, cmap

def clev_cmap_rain_1h_error():
    path = pkg_resources.resource_filename('meteva', "resources/colormaps/color_rain_1h_error.txt")
    cmap, clevs = get_cmap_and_clevs_from_file(path)
    print("不再推荐使用该函数，推荐使用meb.def_cmap_clevs相应功能,请参考color工具中相关说明")
    return clevs, cmap
def clev_cmap_rain_3h_error():
    path = pkg_resources.resource_filename('meteva', "resources/colormaps/color_rain_3h_error.txt")
    cmap, clevs = get_cmap_and_clevs_from_file(path)
    print("不再推荐使用该函数，推荐使用meb.def_cmap_clevs相应功能,请参考color工具中相关说明")
    return clevs, cmap
def clev_cmap_rh_error():
    path = pkg_resources.resource_filename('meteva', "resources/colormaps/color_rh_error.txt")
    cmap, clevs = get_cmap_and_clevs_from_file(path)
    print("不再推荐使用该函数，推荐使用meb.def_cmap_clevs相应功能,请参考color工具中相关说明")
    return clevs, cmap
def clev_cmap_vis_error():
    path = pkg_resources.resource_filename('meteva', "resources/colormaps/color_vis_error.txt")
    cmap, clevs = get_cmap_and_clevs_from_file(path)
    print("不再推荐使用该函数，推荐使用meb.def_cmap_clevs相应功能,请参考color工具中相关说明")
    return clevs, cmap
def clev_cmap_wind_speed_error():
    path = pkg_resources.resource_filename('meteva', "resources/colormaps/color_wind_speed_error.txt")
    cmap, clevs = get_cmap_and_clevs_from_file(path)
    print("不再推荐使用该函数，推荐使用meb.def_cmap_clevs相应功能,请参考color工具中相关说明")
    return clevs, cmap
def clev_cmap_cloud_total_error():
    path = pkg_resources.resource_filename('meteva', "resources/colormaps/color_tcdc_error.txt")
    cmap, clevs = get_cmap_and_clevs_from_file(path)
    print("不再推荐使用该函数，推荐使用meb.def_cmap_clevs相应功能,请参考color工具中相关说明")
    return clevs, cmap


def cmap_clevs_temper_2m_error():
    clevs1 = [-20,-12,-8,-6,-4,-2,-1]
    nclev = len(clevs1)
    colors0 = cm.get_cmap("winter", nclev)
    colors_list = []
    for i in range(nclev):
        colors_list.append(colors0(i))

    clevs2 = [0,1,2,4,6,8,12,20]
    nclev = len(clevs2)
    colors0 = cm.get_cmap("autumn", nclev)
    for i in range(nclev):
        colors_list.append(colors0(nclev - 1 - i))
    clevs1.extend(clevs2)
    cmap = colors.ListedColormap(colors_list, 'indexed')
    return cmap, clevs1


def cmap_clevs_environment():
    clevs1 =[0,35,75,115,150,250,350]
    nclev = len(clevs1)
    colors_list = np.array([[0,228,0],[255,255,0],[255,126,0],[255,0,0],[153,0,76],[126,0,35]])/256
    cmap = colors.ListedColormap(colors_list, 'indexed')
    return cmap, clevs1

def get_cmap_and_clevs_from_file(path):
    clev_cmap = np.loadtxt(path)
    clevs = clev_cmap[:, 0]
    cmap = clev_cmap[:, 1:] / 255
    cmap = cmap.tolist()
    cmap = colors.ListedColormap(cmap, 'indexed')
    return cmap,clevs

def cmap_clevs_bias_(vmax):
    if vmax is None:
        print("设置bias的cmap时需要指定vmax")
    blue = np.array([0, 0, 255]) / 255
    white = np.array([255, 255, 255]) / 255
    red = np.array([255, 0, 0]) / 255
    black = np.array([0, 0, 0]) / 255
    clev_list = [0]
    cmap_list = [blue]
    for v in range(1,6):
        clev_list.append(v * 0.2)
        cmap_list.append(blue * (1 - v * 0.2) + white * v * 0.2)
    for v in range(1,6):
        clev_list.append(1 + v * 0.2)
        cmap_list.append(white * (1 - v * 0.2) + red * v * 0.2)

    for value in range(2, int(vmax + 1), 1):
        clev_list.append(value)
        cmap_list.append((red * (vmax - value) + black * (value - 2)) / (vmax - 2))
    cmap = colors.ListedColormap(cmap_list, 'indexed')
    return cmap,clev_list


def cmap_clevs_bias(vmax):
    if vmax is None:
        print("设置bias的cmap时需要指定vmax")
    blue = np.array([0, 0, 255]) / 255
    white = np.array([255, 255, 255]) / 255
    yellow = np.array([0.7,0.7,0])
    red = np.array([255, 0, 0]) / 255
    pink = np.array([0.5,0,0.5])
    black = np.array([0, 0, 0]) / 255
    clev_list = [0]
    cmap_list = [blue]
    for v in range(1,6):
        clev_list.append(v * 0.2)
        cmap_list.append(blue * (1 - v * 0.2) + white * v * 0.2)

    for v in range(1,6):
        clev_list.append(1 + v * 0.2)
        cmap_list.append(white * (1 - v * 0.2) + yellow * v * 0.2)

    for v in range(1,6):
        clev_list.append(2 + v * 0.2)
        cmap_list.append(yellow * (1 - v * 0.2) + red * v * 0.2)

    for v in range(1,10):
        clev_list.append(3 + v * 0.5)
        cmap_list.append(red * (1 - v * 0.1) + pink * v * 0.1)

    for value in range(8, int(vmax + 1), 1):
        clev_list.append(value)
        cmap_list.append((pink * (vmax - value) + black * (value - 2)) / (vmax - 2))
    cmap = colors.ListedColormap(cmap_list, 'indexed')
    return cmap,clev_list


def cmap_clevs_temper_error_br(vmax):
    blue = np.array([0, 0, 255]) / 255
    white = np.array([255, 255, 255]) / 255
    red = np.array([255, 0, 0]) / 255
    black = np.array([0, 0, 0]) / 255
    clev_list = []
    cmap_list = []

    vmax = int(math.ceil(vmax))

    if vmax > 5:
        for value in range(-vmax, -5):
            clev_list.append(value)
            cmap_list.append((blue * (value + vmax) + black * (-5 - value)) / (vmax -5))

    for v in range(-5,0):
        clev_list.append(v)
        cmap_list.append((blue * (-v) + white * (5+v))/5)
    for v in range(0,6):
        clev_list.append(v)
        cmap_list.append((white * (5 - v) + red*v)/5)

    if vmax>5:
        for value in range(6, vmax):
            clev_list.append(value)
            cmap_list.append((red * (vmax - value) + black * (value-5)) / (vmax - 5))

    cmap = colors.ListedColormap(cmap_list, 'indexed')
    return cmap,clev_list


def cmap_clevs_me(vmin,vmax):
    max_abs = max(abs(vmax),abs(vmin))
    vmax = max_abs
    vmin = -max_abs
    dif = (vmax - vmin) / 10.0
    inte = math.pow(10, math.floor(math.log10(dif)));
    # 用基本间隔，将最大最小值除于间隔后小数点部分去除，最后把间隔也整数化
    r = dif / inte
    if r < 3 and r >= 1.5:
        inte = inte * 2
    elif r < 4.5 and r >= 3:
        inte = inte * 4
    elif r < 5.5 and r >= 4.5:
        inte = inte * 5
    elif r < 7 and r >= 5.5:
        inte = inte * 6
    elif r >= 7:
        inte = inte * 8

    inte = inte/2

    vmin = inte * ((int)(vmin / inte) - 1)

    vmax = inte * ((int)(vmax / inte) + 2)
    clevs1 = np.arange(vmin, -1e-6, inte)
    nclev = len(clevs1)
    colors0 = cm.get_cmap("winter", nclev)
    colors_list = []
    for i in range(nclev):
        colors_list.append(colors0(i))

    clevs2 = np.arange(0, vmax, inte)
    nclev = len(clevs2)
    colors0 = cm.get_cmap("autumn", nclev)
    for i in range(nclev):
        colors_list.append(colors0(nclev -1 - i))
    clevs = np.arange(vmin, vmax, inte)
    cmap = colors.ListedColormap(colors_list, 'indexed')
    return cmap,clevs


def cmap_clevs_me_w0(vmin,vmax):
    max_abs = max(abs(vmax),abs(vmin))
    vmax = max_abs
    vmin = -max_abs
    dif = (vmax - vmin) / 10.0
    inte = math.pow(10, math.floor(math.log10(dif)));
    # 用基本间隔，将最大最小值除于间隔后小数点部分去除，最后把间隔也整数化
    r = dif / inte
    if r < 3 and r >= 1.5:
        inte = inte * 2
    elif r < 4.5 and r >= 3:
        inte = inte * 4
    elif r < 5.5 and r >= 4.5:
        inte = inte * 5
    elif r < 7 and r >= 5.5:
        inte = inte * 6
    elif r >= 7:
        inte = inte * 8

    inte = inte/2

    vmin = inte * int(math.floor((vmin / inte)))

    vmax = inte * ((int)(vmax / inte) + 1)

    clevs = np.arange(vmin, vmax, inte)
    num = len(clevs)
    rgb_colors = []
    sp = 3/4
    step = 360*sp / num
    #print(num)
    delta0 = int((num-1)/2)
    left = delta0 % 4
    for i in range(num):
        delta = abs(i-((num-1)/2))
        if i < (num-1)/2:
            h = 0.25 +(1-sp)/2 + delta * step / 360  # 首先均匀的取不同的色相，保持色相维度的差异最大化
            if delta <=(left + 4):
                if delta <=1:delta = 0
                s = 0.75 + 0.25 * delta / (left + 4)  # 通过一个折线波浪 设置不同的饱和度
                l = 0.9 - 0.6 * np.power(delta / (left + 4),1)  # 通过一个折线波浪 设置不同的亮度
            else:
                i1 = i % 4
                s = 0.75 + 0.25 * (i1 / 4)  # 通过一个折线波浪 设置不同的饱和度
                l = 0.3 + 0.6 * np.power((i1 / 4),1)  # 通过一个折线波浪 设置不同的亮度

        else:
            h = 0.25 - (1 - sp) / 2 - delta * step / 360  # 首先均匀的取不同的色相，保持色相维度的差异最大化
            if delta < (left + 4):
                s = 0.75 + 0.25 * delta / (left + 4)  # 通过一个折线波浪 设置不同的饱和度
                l = 0.9 - 0.6 * np.power(delta / (left + 4),1)  # 通过一个折线波浪 设置不同的亮度
            else:
                i1 = 4 - (num-1 - i) % 4
                s = 1 - 0.25 * i1 / 4  # 通过一个折线波浪 设置不同的饱和度
                l = 0.75 - 0.35 * np.power(i1 / 4,1)  # 通过一个折线波浪 设置不同的亮度



        rgb1 = colorsys.hls_to_rgb(h, l, s)
        rgb_colors.append(rgb1)

    clevs = np.arange(vmin, vmax, inte)
    cmap = colors.ListedColormap(rgb_colors, 'indexed')
    return cmap,clevs

def cmap_clevs_me_new(vmin,vmax):
    max_abs = max(abs(vmax),abs(vmin))
    vmax = max_abs
    vmin = -max_abs
    dif = (vmax - vmin) / 10.0
    inte = math.pow(10, math.floor(math.log10(dif)));
    # 用基本间隔，将最大最小值除于间隔后小数点部分去除，最后把间隔也整数化
    r = dif / inte
    if r < 3 and r >= 1.5:
        inte = inte * 2
    elif r < 4.5 and r >= 3:
        inte = inte * 4
    elif r < 5.5 and r >= 4.5:
        inte = inte * 5
    elif r < 7 and r >= 5.5:
        inte = inte * 6
    elif r >= 7:
        inte = inte * 8

    inte = inte/2

    vmin = inte * ((int)(vmin / inte) - 1)

    vmax = inte * ((int)(vmax / inte) + 2)
    clevs1 = np.arange(vmin, -1e-6, inte)
    nclev = len(clevs1)
    colors0 = cm.get_cmap("winter", nclev)
    colors_list = []
    for i in range(nclev):
        if i == nclev - 1:
            c1 = np.array(list(colors0(i)))/5 +4/5
            colors_list.append(c1)
        else:
            colors_list.append(colors0(i))

    clevs2 = np.arange(0, vmax, inte)
    nclev = len(clevs2)
    colors0 = cm.get_cmap("autumn", nclev)
    for i in range(nclev):
        if i ==0:
            c1 = np.array(list(colors0(nclev -1 - i)))/5 +4/5
            colors_list.append(c1)
        else:
            colors_list.append(colors0(nclev -1 - i))

    clevs = np.arange(vmin, vmax, inte)
    cmap = colors.ListedColormap(colors_list, 'indexed')
    return cmap,clevs


def cmap_clevs_mae(vmax):

    dif = (vmax) / 10.0
    inte = math.pow(10, math.floor(math.log10(dif)));
    # 用基本间隔，将最大最小值除于间隔后小数点部分去除，最后把间隔也整数化
    r = dif / inte
    if r < 3 and r >= 1.5:
        inte = inte * 2
    elif r < 4.5 and r >= 3:
        inte = inte * 4
    elif r < 5.5 and r >= 4.5:
        inte = inte * 5
    elif r < 7 and r >= 5.5:
        inte = inte * 6
    elif r >= 7:
        inte = inte * 8

    inte = inte / 2

    vmin = 0

    vmax = inte * ((int)(vmax / inte) + 1)

    clevs = np.arange(vmin, vmax, inte)
    num = len(clevs)
    rgb_colors = []

    for i in range(num):
        h = 0.8 - i/num
        s = 1
        l = 1 - 0.6 * (i+1)/num
        rgb1 = colorsys.hls_to_rgb(h, l, s)
        rgb_colors.append(rgb1)

    clevs = np.arange(vmin, vmax, inte)
    cmap = colors.ListedColormap(rgb_colors, 'indexed')
    return cmap, clevs


def cmap_clevs_me_bwr(vmin,vmax):

    # max_abs = math.ceil(max(abs(vmax),abs(vmin)))
    # inte = 1
    # vmin = inte * ((int)(vmin / inte))
    #
    max_abs = max(abs(vmax),abs(vmin))
    vmax = max_abs
    vmin = -max_abs
    dif = (vmax - vmin) / 10.0
    inte = math.pow(10, math.floor(math.log10(dif)));
    # 用基本间隔，将最大最小值除于间隔后小数点部分去除，最后把间隔也整数化
    r = dif / inte
    if r < 3 and r >= 1.5:
        inte = inte * 2
    elif r < 4.5 and r >= 3:
        inte = inte * 4
    elif r < 5.5 and r >= 4.5:
        inte = inte * 5
    elif r < 7 and r >= 5.5:
        inte = inte * 6
    elif r >= 7:
        inte = inte * 8

    inte = inte/2

    vmin = inte * ((int)(vmin / inte) - 1)
    vmax = inte * ((int)(vmax / inte) + 2)
    max_abs_h = math.ceil(max(abs(vmax),abs(vmin)))/2
    clevs = []
    colors_list = []

    for i in np.arange(vmin,0,inte).tolist():
        clevs.append(i)
        if i <= -max_abs_h:
            rgb = [0,0,(i - vmin)/max_abs_h]
        else:
            rgb = [1+ i/max_abs_h,1+ i/max_abs_h,1]
        #print(rgb)
        if i>=-inte:
            rgb= [1,1,1]
        colors_list.append(rgb)

    for i in np.arange(0,vmax,inte).tolist():
        clevs.append(i)
        if i <= max_abs_h:
            rgb = [1,1- i/max_abs,1- i/max_abs]
        else:
            rgb = [(vmax - i)/max_abs_h,0,0]
        if i < inte:
            rgb = [1,1,1]
        colors_list.append(rgb)
    #print(colors_list)
    cmap = colors.ListedColormap(colors_list, 'indexed')
    return cmap,clevs


def cmap_clevs_mode(vmax):
    '''
    定义mode目标绘制的colorbar
    :param vmax:  最大的目标编号
    :return:
    '''
    if vmax <12:
        cmap1,clevs1 = def_cmap_clevs(cmap="Paired",clevs=np.arange(1,vmax+1))
    elif vmax <20:
        cmap1, clevs1 = def_cmap_clevs(cmap="tab20b", clevs=np.arange(1, vmax+1))
    else:
        cmap1, clevs1 = def_cmap_clevs(cmap="gist_rainbow", clevs=np.arange(1, vmax+1))

    #cmap2 = colors.ListedColormap([[80,80,80],[125,125,125],[255,255,255]], 'indexed')
    #clevs2 = [-2,-1,0]
    cmap2,clevs2 = def_cmap_clevs(cmap = "gray",clevs = [-2,-1,0],vmin =-1,vmax = 0,cut_colorbar=True)
    cmap2,clevs2 = get_part_cmap_and_clevs(cmap2,clevs2,vmin= -0.5,vmax = 0,cut_accurate=True)
    cmap3,clevs3 = merge_cmap_clevs(cmap2,clevs2,cmap1,clevs1)
    clevs3 = (np.array(clevs3) -0.5).tolist()

    #show_cmap_clev(cmap3,clevs3)
    return cmap3,clevs3

def cmap_clevs_ts():
    clevs = np.arange(0,1.01,0.1)
    nclev = len(clevs)
    colors0 = cm.get_cmap("jet", nclev)
    colors_list = []
    for i in range(nclev):
        clev = clevs[i]
        cl = []
        cl[0:3] = colors0(i)[0:3]
        cl.append(clev)
        colors_list.append(tuple(cl))
        #print(colors0(i))
    cmap = colors.ListedColormap(colors_list, 'indexed')
    return cmap,clevs

def hour():
    #import matplotlib.colors as colors
    cmap = np.array([[0,0,255],
                     [0,115,255],
                     [16,190,56],
                     [255,255,0],
                     [254,147,5],
                     [255,18,0],
                     [177,0,62],
                     [117,0,121],
                     [0,0,255],])/255
    cmap = cmap.tolist()
    cmap = colors.ListedColormap(cmap, 'indexed')
    clevs = np.arange(0,24,3).tolist()
    return  cmap ,clevs


def cmap_clevs_radar():
    clevs = np.arange(5,75,5).tolist()
    clevs1 = [-5,0]
    clevs1.extend(clevs)
    colors_list = np.array([
        [255, 255, 255],
        [0,0,246],
        [1,160,246],
        [78,242,242],
        [1,255,0],
        [0,200,0],
        [1,144,0],
        [255,255,0],
        [231,192,0],
        [255,144,0],
        [255,0,0],
        [214,0,0],
        [192,0,0],
        [255,0,240],
        [120,0,132],
        [173,144,240]
    ])/256
    cmap = colors.ListedColormap(colors_list, 'indexed')
    return cmap,clevs1

def cmap_clevs_far():
    clevs = np.arange(0,1.01,0.1)
    nclev = len(clevs)
    colors0 = cm.get_cmap("jet", nclev)
    colors_list = []
    for i in range(nclev):
        colors_list.append(colors0(i))
    cmap = colors.ListedColormap(colors_list, 'indexed')
    return cmap,clevs

def cmap_clevs_mr():
    clevs = np.arange(0,1.01,0.1)
    nclev = len(clevs)
    colors0 = cm.get_cmap("jet", nclev)
    colors_list = []
    for i in range(nclev):
        colors_list.append(colors0(i))
    cmap = colors.ListedColormap(colors_list, 'indexed')
    return cmap,clevs


def cmap_clevs_error(vmin,vmax):

    max_abs = max(abs(vmax),abs(vmin))
    vmax = max_abs
    vmin = -max_abs
    dif = (vmax - vmin) / 10.0
    inte = math.pow(10, math.floor(math.log10(dif)));
    # 用基本间隔，将最大最小值除于间隔后小数点部分去除，最后把间隔也整数化
    r = dif / inte
    if r < 3 and r >= 1.5:
        inte = inte * 2
    elif r < 4.5 and r >= 3:
        inte = inte * 4
    elif r < 5.5 and r >= 4.5:
        inte = inte * 5
    elif r < 7 and r >= 5.5:
        inte = inte * 6
    elif r >= 7:
        inte = inte * 8
    vmin = inte * ((int)(vmin / inte) - 1)
    vmax = inte * ((int)(vmax / inte) + 2)
    clevs = np.arange(vmin, vmax, inte)
    nclev = len(clevs)
    colors0 = cm.get_cmap("bwr", nclev)
    colors_list = []
    for i in range(nclev):
        colors_list.append(colors0(i))
    cmap = colors.ListedColormap(colors_list, 'indexed')
    return cmap,clevs


def get_steps_range(line):
    num = len(line)
    max_nstep = 0

    #首先获得line中各段的起始坐标和长度
    start_list = [0]
    lenght_list = []
    n = 0
    for k in range(1,num):
        n += 1
        if line[k] != line[k-1]:
            lenght_list.append(n)
            start_list.append(k)
            n = 0
    lenght_list.append(n)

    #统计长度序列中保持不变的部分
    start_list_length = [0]
    lenght_list_lenght = []
    n = 0
    for k in range(1,len(lenght_list)):
        n += 1
        if abs(lenght_list[k] - lenght_list[k-1]) > 1:
            lenght_list_lenght.append(n)
            start_list_length.append(k)
            n = 0
    lenght_list_lenght.append(n)

    #计算总共有多少个color等级
    step_num  =  np.max(np.array(lenght_list_lenght))
    #计算等长序列最大个数在lenght_list_lenght对应的起始位置
    k = np.argmax(np.array(lenght_list_lenght))

    # 计算lenght_list_lenght起始坐标在lenght_list中的位置
    j = start_list_length[k]

    # lenght_list中的位置在整个line中的位置
    i_start_list = []
    for i in range(step_num + 1):
        i_start_list.append(start_list[j + i])

    return step_num,i_start_list

def get_cmap_from_picture(path,show = False):

    #im的第一维是y方向，第二维是x方向，第三维是rgb
    im = image.imread(path)
    #首先沿着x方向寻找
    rgb_to_1d = np.zeros((im.shape[0],im.shape[1]))
    rgb_to_1d = im[:,:,0] * 256* 256 + im[:,:,1] * 256 + im[:,:,2]


    color_type_num_y = np.zeros(im.shape[0])
    for y in range(im.shape[0]):
        line = rgb_to_1d[y,:] #line 是一条沿着x方向的线
        rgb_list = line.tolist()
        color_set  = set(rgb_list)
        color_type_num_y[y] = len(color_set)

    color_type_num_x = np.zeros(im.shape[1])
    for x in range(im.shape[1]):
        line = rgb_to_1d[:,x] #line_x 是一条沿着y方向的线
        rgb_list = line.tolist()
        color_set  = set(rgb_list)
        color_type_num_x[x] = len(color_set)


    max_color_type_y = np.max(color_type_num_y)
    max_color_type_x = np.max(color_type_num_x)
    max_color_type = max(max_color_type_x,max_color_type_y)


    max_step_num = 0
    i_start = 0
    i_end = 0
    j_start  = 0
    j_end = 0
    color_list = []

    ij_list = np.where(color_type_num_x == max_color_type)[0]
    if(len(ij_list) > 0):
        k_list_list = []
        k_list = [0]
        for k in range(1,len(ij_list)):
            if ij_list[k] - ij_list[k-1] == 1:
                k_list.append(k)
            else:
                k_list_list.append(k_list)
                k_list = [k]
        k_list_list.append(k_list)

        for k in range(len(k_list_list)):
            k_list =k_list_list[k]
            mid = int((ij_list[0] + ij_list[-1])/2)
            line = rgb_to_1d[:,mid]
            step_num,start_list = get_steps_range(line)
            if step_num > max_step_num:
                max_step_num = step_num
                i_start = ij_list[0]
                i_end = ij_list[-1]
                j_start = start_list[0]
                j_end = start_list[-1]
                color_list = []
                for i in range(step_num):
                    start = start_list[-i-2]
                    color_list.append(im[start,mid,:])


    #
    ij_list = np.where(color_type_num_y == max_color_type)[0]
    if len(ij_list) >0:
        k_list_list = []
        k_list = [0]
        for k in range(1,len(ij_list)):
            if ij_list[k] - ij_list[k-1] == 1:
                k_list.append(k)
            else:
                k_list_list.append(k_list)
                k_list = [k]
        k_list_list.append(k_list)

        for k in range(len(k_list_list)):
            k_list =k_list_list[k]
            mid = int((ij_list[0] + ij_list[-1])/2)
            line = rgb_to_1d[mid,:]
            step_num,start_list = get_steps_range(line)
            if step_num > max_step_num:
                max_step_num = step_num
                j_start = ij_list[0]
                j_end = ij_list[-1]
                i_start = start_list[0]
                i_end = start_list[-1]
                color_list = []
                for i in range(step_num):
                    start = start_list[i]
                    color_list.append(im[mid,start,:])

    cmap = colors.ListedColormap(color_list, 'indexed')

    if show:
        show_cmap_clev(cmap)

    return cmap


def get_cmap_and_clevs_by_element_name(element_name):
    path = None
    if element_name == "temp":
        path = pkg_resources.resource_filename('meteva', "resources/colormaps/color_temp_2m.txt")
    elif element_name == "rain_1h":
        path = pkg_resources.resource_filename('meteva', "resources/colormaps/color_rain_1h.txt")
    elif element_name == "rain_1h_error":
        path = pkg_resources.resource_filename('meteva', "resources/colormaps/color_rain_1h_error.txt")
    elif element_name == "rain_3h":
        path = pkg_resources.resource_filename('meteva', "resources/colormaps/color_rain_3h.txt")
    elif element_name == "rain_3h_error":
        path = pkg_resources.resource_filename('meteva', "resources/colormaps/color_rain_3h_error.txt")
    elif element_name == "rh":
        path = pkg_resources.resource_filename('meteva', "resources/colormaps/color_rh.txt")
    elif element_name == "rh_error":
        path = pkg_resources.resource_filename('meteva', "resources/colormaps/color_rh_error.txt")
    elif element_name == "vis":
        path = pkg_resources.resource_filename('meteva', "resources/colormaps/color_vis.txt")
    elif element_name == "vis_error":
        path = pkg_resources.resource_filename('meteva', "resources/colormaps/color_vis_error.txt")
    elif element_name == "wind_speed":
        path = pkg_resources.resource_filename('meteva', "resources/colormaps/color_wind_speed.txt")
    elif element_name == "wind_speed_error":
        path = pkg_resources.resource_filename('meteva', "resources/colormaps/color_wind_speed_error.txt")
    elif element_name == "tcdc":
        path = pkg_resources.resource_filename('meteva', "resources/colormaps/color_tcdc.txt")
    elif element_name == "tcdc_error":
        path = pkg_resources.resource_filename('meteva', "resources/colormaps/color_tcdc_error.txt")
    cmap,clevs = get_cmap_and_clevs_from_file(path)
    return cmap,clevs

def get_part_cmap_and_clevs(cmap_all,clev_all,vmax,vmin,cut_accurate = False):


    if cut_accurate:
        start_i = 0
        for i in range(len(clev_all) - 1):
            if vmin < clev_all[i + 1]:
                start_i = i
                break
        end_i = 0
        for i in range(len(clev_all) - 1):
            if vmax > clev_all[i]:
                end_i = i+2

    else:
        if len(clev_all) < 20:
            return cmap_all, clev_all
        start_i = 0
        for i in range(len(clev_all)-1):
            if vmin<clev_all[i+1]:
                start_i = i
                break
        end_i = 0
        for i in range(len(clev_all)-1):
            if vmax > clev_all[i]:
                end_i = i+2
        if end_i - start_i<=15:
            end_i = start_i+15


    clevs_part = clev_all[start_i:end_i]
    if hasattr(cmap_all,"colors"):
        cmap_colors = cmap_all.colors
        cmap_colors_part = cmap_colors[start_i:end_i]
        cmap_part = colors.ListedColormap(cmap_colors_part, 'indexed')
    else:
        cmap_part = cmap_all
    return cmap_part,clevs_part

def write_cmap_and_clevs(cmap,clevs,path):
    num = len(clevs)
    clev_cmap = np.zeros((num,4))
    clev_cmap[:,0] = clevs[:]
    cmap_data = np.array(cmap.colors)
    max_data = np.max(cmap_data)
    if max_data <=1:
        clev_cmap[:,1:] = cmap_data * 255
    else:
        clev_cmap[:, 1:] = cmap_data
    np.savetxt(path,clev_cmap,fmt = "%f")

def get_denser_cmap(cmap,multi_num):
    if hasattr(cmap,"colors"):
        colors0 = np.array(cmap.colors)
        colors_list = []
        num = len(colors0) - 1
        for i in range(num):
            for j in range(multi_num):
                color1 = (colors0[i,:]*(multi_num-j) + colors0[i+1,:] * j)/multi_num
                colors_list.append(color1.tolist())
        colors_list.append(colors0[-1])
        cmap_denser = colors.ListedColormap(colors_list, 'indexed')
        return cmap_denser
    else:
        return cmap

def show_cmap_clev(cmap,clev = None):
    """
    Show color map.
    :param cmap: color map instance.
    :return: None
    """
    n_colors = len(cmap.colors)
    width = 8
    heigh = 2
    n_h = int(heigh * n_colors/width)
    im = np.outer(np.ones(n_h), np.arange(n_colors))


    fig, ax = plt.subplots(1, figsize=(width, heigh),
                           subplot_kw=dict(xticks=[], yticks=[]))
    if clev is not None:
        max_tick = 10
        step = int(math.ceil(n_colors/max_tick))
        x = np.arange(0,n_colors,step).astype(np.int32)
        #print(x)
        ax.set_xticks(x)
        labels = []
        for i in range(x.size):
            labels.append(round(clev[x[i]],6))
        ax.set_xticklabels(labels)
    ax.imshow(im, cmap=cmap)


def get_color_list(legend_num):
    colors_list = []
    if legend_num<=10:
        colors = cm.get_cmap("tab10")
        for i in range(legend_num):
            colors_list.append(colors(i))
    elif legend_num <=20:
        colors = cm.get_cmap("tab20")
        for i in range(legend_num):
            colors_list.append(colors(i))
    else:
        colors = cm.get_cmap('gist_rainbow', 128)
        for i in range(legend_num):
            color_grade = i / legend_num
            colors_list.append(colors(color_grade))
    return colors_list


def get_cmap_and_clevs_by_name(cmap_name,vmin,vmax):
    path = pkg_resources.resource_filename('meteva', "resources/colormaps/color_"+cmap_name+".txt")
    if os.path.exists(path):
        cmap,clevs = get_cmap_and_clevs_from_file(path)
    elif cmap_name == "bias":
        cmap,clevs = cmap_clevs_bias(vmax)
    elif cmap_name == "error":
        cmap,clevs = cmap_clevs_error(vmin,vmax)
    elif cmap_name == "me":
        cmap,clevs = cmap_clevs_me(vmin,vmax)
    elif cmap_name == "me_bwr":
        cmap,clevs = cmap_clevs_me_bwr(vmin,vmax)
    elif cmap_name == "me_w0":
        cmap,clevs = cmap_clevs_me_w0(vmin,vmax)
    elif cmap_name == "ts":
        cmap,clevs= cmap_clevs_ts()
    elif cmap_name == "far":
        cmap,clevs = cmap_clevs_far()
    elif cmap_name == "far":
        cmap,clevs = cmap_clevs_far()
    elif cmap_name == "mode":
        cmap, clevs = cmap_clevs_mode(vmax)
    elif cmap_name == "temper_2m_error":
        cmap, clevs = cmap_clevs_temper_2m_error()
    elif cmap_name == "temper_error_br":
        cmap, clevs = cmap_clevs_temper_error_br(vmax)
    elif cmap_name =="environment":
        cmap,clevs = cmap_clevs_environment()
    elif cmap_name =="radar":
        cmap,clevs = cmap_clevs_radar()
    elif cmap_name =="mae":
        cmap, clevs = cmap_clevs_mae(vmax)
    elif cmap_name =="hour":
        cmap,clevs = hour()
    else:
        print("该配色方案名称不识别")
        return None,None
    return cmap,clevs


class cmaps:
    rain_1h = "rain_1h"
    rain_1h_error = "rain_1h_error"
    rain_3h = "rain_3h"
    rain_3h_error = "rain_3h_error"
    rain_24h = "rain_24h"
    rain_24h_error = "rain_24h_error"
    temp_2m = "temp_2m"
    temper_2m_error = "temper_2m_error"
    rh = "rh"
    rh_error = "rh_error"
    vis = "vis"
    vis_error = "vis_error"
    wind_speed = "wind_speed"
    wind_speed_error = "wind_speed_error"
    tcdc = "tcdc"
    tcdc_error  = "tcdc_error"
    bias = "bias"
    error = "error"
    ts = "ts"
    far = "far"
    mr = "mr"
    mode = "mode"
    me = "me"
    temper_error_br ="temper_error_br"
    environment = "environment"
    radar = "radar"
    me_bwr = "me_bwr"
    me_w0 = "me_w0"
    mae = "mae"
    hour = "hour"


def coordinate_cmap_to_clevs(cmap,clevs):
    if hasattr(cmap, "colors"):
        colors0 = np.array(cmap.colors)
        colors_list = []
        ncmap = len(colors0)
        nclev = len(clevs)
        if nclev <3:
            print("clevs' size must bigger than 1")
        for i in range(nclev):
            if nclev>1:
                j = i * (ncmap-1) /(nclev-1)
            else:
                j = 0
            j0 = int(j)
            j1 = min(j0 + 1,ncmap-1)
            dj = j - j0
            color1 = (colors0[j0, :] * (1 - dj) + colors0[j1, :] * dj)
            colors_list.append(color1.tolist())
        cmap_co = colors.ListedColormap(colors_list, 'indexed')
        return cmap_co,clevs
    else:
        print("cmap is not colormap")
        return



def def_cmap_clevs(cmap = "rainbow",clevs = None,vmin = None,vmax = None,cut_colorbar = True):
    #  # 判断是meteva自定义的颜色类型，这从meteva资源文件或函数里生成cmap1 和clevs1
    clevs1 = None
    cmap1 = None

    if isinstance(cmap,str):
        cmap_class = cmaps()
        if hasattr(cmap_class, cmap):
            if vmin is None:
                vmin = np.min(np.array(clevs))
            if vmax is None:
                vmax = np.max(np.array(clevs))
            if vmax is not None and vmin == vmax:vmin = vmax - 1
            cmap,clevs1= get_cmap_and_clevs_by_name(cmap, vmin, vmax)
    if isinstance(cmap,list):
        if isinstance(cmap[0],list):
            if isinstance(clevs,list):
                if len(clevs)==len(cmap)+1:
                    cmap.append(cmap[-1])
            cmap_array= np.array(cmap)
            if np.max(cmap_array)>1:
                cmap_array =cmap_array / 256
            cmap_list = cmap_array.tolist()
        else:
            cmap_list = []
            for c1 in cmap:
                r1 = int(c1[1:3].upper(), 16)/256
                g1 = int(c1[3:5].upper(), 16)/256
                b1 = int(c1[5:7].upper(), 16)/256
                cmap_list.append([r1,g1,b1])
        cmap = colors.ListedColormap(cmap_list, 'indexed')


    #设置clevs2
    if clevs is None:
        if clevs1 is None:
            # 如果cmap字符不是meteva内集成的，则判断它是matplotlib集成的，则进一步根据最大最小值生成cmap，clevs
            if vmin is None or vmax is None:
                print("clev and vmin/vmax cann't be None at the same time while cmap is matplotlib strType cmap")
                return
            if vmax - vmin < 1e-10:
                vmax = vmin + 1.1
            dif = (vmax - vmin) / 10.0
            inte = math.pow(10, math.floor(math.log10(dif)));
            # 用基本间隔，将最大最小值除于间隔后小数点部分去除，最后把间隔也整数化
            r = dif / inte
            if r < 3 and r >= 1.5:
                inte = inte * 2
            elif r < 4.5 and r >= 3:
                inte = inte * 4
            elif r < 5.5 and r >= 4.5:
                inte = inte * 5
            elif r < 7 and r >= 5.5:
                inte = inte * 6
            elif r >= 7:
                inte = inte * 8


            vmin = inte * (math.floor(vmin / inte))
            vmax = inte * (math.ceil(vmax / inte)+0.5)
            clevs2 = np.arange(vmin, vmax, inte)
        else:
            clevs2 = clevs1
    else:
        clevs2 = clevs

    # 设置cmap2
    if cmap1 is None:
        if isinstance(cmap,str):
            nclev = len(clevs2)
            colors0 = cm.get_cmap(cmap, nclev)
            colors_list = []
            for i in range(nclev):
                colors_list.append(colors0(i))
            cmap2 = colors.ListedColormap(colors_list, 'indexed')
        else:
            cmap2 = cmap
    else:
        cmap2 = cmap1


    #将cmap2 和clev2 协调
    cmap3,clevs3 = coordinate_cmap_to_clevs(cmap2,clevs2)

    # 从cmap3 和cmap3中提取部分colorbar

    if vmin is not None and vmax is not None and cut_colorbar:
        cmap4,clevs4 = get_part_cmap_and_clevs(cmap3, clevs3, vmax, vmin)
    else:
        cmap4,clevs4  = cmap3,clevs3

    return cmap4,clevs4


def merge_cmap_clevs(cmap0,clevs0,cmap1,clevs2):
    '''
    合并两个colorbar
    :param cmap0:
    :param clevs0:
    :param cmap1:
    :param clevs2:
    :return:
    '''
    colors0 = np.array(cmap0.colors).tolist()
    colors1 = np.array(cmap1.colors).tolist()
    colors0.extend(colors1)
    cmap_m = colors.ListedColormap(colors0, 'indexed')
    clevs0.extend(clevs2)
    return cmap_m,clevs0


def cart2sph(x, y, z):
    XsqPlusYsq = x ** 2 + y ** 2
    r = math.sqrt(XsqPlusYsq + z ** 2)  # r
    lat = math.atan2(z, math.sqrt(XsqPlusYsq))  # theta
    lon = math.atan2(y, x)  # phi
    return r, lat, lon


def get_seprated_rgb_method1(num):
    '''
    通过再hsl球体空间种，寻找距离最远的num个点来，获得对比度最大的num个颜色
    :param num:  需要返回的颜色数目
    :return:  rgb列表
    '''
    loc0 = np.zeros((3, num + 1))
    loc0[:, 0] = [0, 0, 1]  # 固定第一个点是白色
    loc0[:, 1] = [0, 1, 0]  # 固定第二个点是红色
    loc0[0, 2:] = np.arange(1, num) / num  #初始化其他点的位置

    # 将每个颜色看作hsl空间种的一个同号带电粒子，根据每个受力调整其位置，逐步迭代方法，使它们距离更远。
    for i in range(20000):
        # 计算受力
        force = np.zeros((3, num))
        for j in range(1, num):
            other = np.zeros((3, num))
            other[:, 0:j + 1] = loc0[:, 0:j + 1]
            other[:, j + 1:] = loc0[:, j + 2:]
            this = loc0[:, j + 1]
            dxyz = this.T - other.T

            dis2 = np.sum(np.power(dxyz, 2), axis=1)
            force_j_m = dxyz.T / dis2.T
            force_j = np.sum(force_j_m, axis=1)
            force[:, j] = force_j[:]
        # 根据受力方向调整每个
        force_s = np.sqrt(np.sum(np.power(force, 2), axis=0))
        mean_f = np.mean(force_s)
        for j in range(1, num):
            if force_s[j] > mean_f:
                move = 0.009 * force[:, j] / force_s[j]
                loc_new = loc0[:, j + 1] + move
                dis0 = np.sqrt(np.sum(np.power(loc_new, 2)))
                if dis0 > 1:
                    loc_new /= dis0
                loc0[:, j + 1] = loc_new
    # 直角坐标转换成球坐标，再转换成hsl色彩模式
    loc0 = loc0.T.tolist()
    loc0.pop(0)
    loc1 = []
    for j in range(num):
        r, lat, lon = cart2sph(loc0[j][0], loc0[j][1], loc0[j][2])
        s = r
        h = lon / (2 * math.pi)
        l = 0.5 + lat / math.pi
        loc1.append([h, l, s])
    loc1 = np.array(loc1)
    loc1[:, 0] -= loc1[0, 0]
    loc2 = loc1[loc1[:, 0].argsort(), :]  # 按照色相排序
    loc3 = np.zeros(loc2.shape)
    index = np.where(loc2[:, 0] < 0)
    nm = len(index[0])
    loc3[num - nm:, :] = loc2[0:nm, :]
    loc3[:num - nm, :] = loc2[nm:, :]  # 将红色调整到第一个

    # 转换成rgb
    rgb_colors = []
    for j in range(num):
        rgb1 = colorsys.hls_to_rgb(loc3[j, 0], loc3[j, 1], loc3[j, 2])
        rgb_colors.append(rgb1)
    return rgb_colors


def get_seprated_rgb_method2(num):
    '''
    设置较大区分度的一组颜色
    :param num:  需要返回的颜色数目
    :return:  rgb列表
    '''
    rgb_colors = []
    step = 360.0 / num
    for i in range(num):
        h = i * step / 360  # 首先均匀的取不同的色相，保持色相维度的差异最大化
        i1 = i % 6
        if i1 <= 2:
            di = 0.5 - i1
        else:
            di = 0.5 + i1 - 4
        s = 0.75 + 0.25 * di / 1.5  # 通过一个折线波浪 设置不同的饱和度
        l = 0.5 + 0.25 * di / 1.5    # 通过一个折线波浪 设置不同的亮度
        rgb1 = colorsys.hls_to_rgb(h, l, s)
        rgb_colors.append(rgb1)

    return rgb_colors


def set_plot_color_dict_method0(member_list):
    if len(member_list) >20:
        print("tab20颜色方案只能支持member_list长度小于等于20的情况")
        return

    cm0 = cm.get_cmap("tab20")
    color_list0 = []
    for i in range(20):
        color_list0.append(cm0(i))
    plot_color_dict = {}
    for i in range(len(member_list)):
        plot_color_dict[member_list[i]] = color_list0[i]
    meteva.base.plot_color_dict = plot_color_dict
    return

def set_plot_color_dict_method1(member_list):
    color_list0 = get_seprated_rgb_method1(len(member_list))
    plot_color_dict = {}
    for i in range(len(member_list)):
        plot_color_dict[member_list[i]] = color_list0[i]
    meteva.base.plot_color_dict = plot_color_dict
    return

def set_plot_color_dict_method2(member_list):
    color_list0 = get_seprated_rgb_method2(len(member_list))
    plot_color_dict = {}
    for i in range(len(member_list)):
        plot_color_dict[member_list[i]] = color_list0[i]
    meteva.base.plot_color_dict = plot_color_dict
    return



import pandas as pd
import meteva

def features_tran(features):
    if "member" not in features.keys():
        print("警告：输入的检验属性数据中没有包含预报名称，请使用升级1.7.5版本以上MetEva重新运行MODE方法输出检验结果后，再检验数据载入和分析")
        return None,None
    try:
        time1 = meteva.base.all_type_time_to_time64(features["time"])
        tabel = features["feature_table"]["contingency_table_yesorno"]

            # hmfc = pd.DataFrame({
            #     "level":[0],"time":[time1],"dtime":[features["dtime"]],"id":[meteva.base.IV],
            #     "lon":[meteva.base.IV],"lat":[meteva.base.IV],
            #     "H":[tabel["Hits"]],
            #     "F":[tabel["False alarms"]],
            #     "M":[tabel["Misses"]],
            #     "C":[tabel["Correct negatives"]]
            # })
        #升级后删除了id，lon，lat列，增加了member列，方便调用透视分析模块
        hmfc = pd.DataFrame({
            "level":[0],"time":[time1],"dtime":[features["dtime"]],"member":[features["member"]],
            "H":[tabel["Hits"]],
            "F":[tabel["False alarms"]],
            "M":[tabel["Misses"]],
            "C":[tabel["Correct negatives"]]
        })
        hmfc = hmfc.sort_values(by = ["level","time","dtime"])
        nmatch = features["match_count"]
        label_list_matched = features["label_list_matched"]
        df_list = []
        for i in label_list_matched:
            fi = features[i]
            dict1 = {
                "level": [0], "time": [time1], "dtime": [features["dtime"]], "id": [i],
                "lon":[fi["feature_props"]["ob"]["centroid"]["x"]],"lat":[fi["feature_props"]["ob"]["centroid"]["y"]],
                "ob_lenghts_MajorAxis":[fi["feature_axis"]["ob"]["lengths"]["MajorAxis"]],
                "ob_lenghts_MinorAxis": [fi["feature_axis"]["ob"]["lengths"]["MinorAxis"]],
                "ob_window_x0":[fi["feature_axis"]["ob"]["window"]["x0"]],
                "ob_window_y0": [fi["feature_axis"]["ob"]["window"]["y0"]],
                "ob_window_x1": [fi["feature_axis"]["ob"]["window"]["x1"]],
                "ob_window_y1": [fi["feature_axis"]["ob"]["window"]["y1"]],
                "ob_centroid_x": [fi["feature_props"]["ob"]["centroid"]["x"]],
                "ob_centroid_y": [fi["feature_props"]["ob"]["centroid"]["y"]],
                "ob_area": [fi["feature_props"]["ob"]["area"]],
                "ob_intensity": [fi["feature_props"]["ob"]["intensity"]],
                "ob_angle": [fi["feature_axis"]["ob"]["OrientationAngle"]["MajorAxis"]],

                "fo_lenghts_MajorAxis": [fi["feature_axis"]["fo"]["lengths"]["MajorAxis"]],
                "fo_lenghts_MinorAxis": [fi["feature_axis"]["fo"]["lengths"]["MinorAxis"]],
                "fo_window_x0": [fi["feature_axis"]["fo"]["window"]["x0"]],
                "fo_window_y0": [fi["feature_axis"]["fo"]["window"]["y0"]],
                "fo_window_x1": [fi["feature_axis"]["fo"]["window"]["x1"]],
                "fo_window_y1": [fi["feature_axis"]["fo"]["window"]["y1"]],
                "fo_centroid_x": [fi["feature_props"]["fo"]["centroid"]["x"]],
                "fo_centroid_y": [fi["feature_props"]["fo"]["centroid"]["y"]],
                "fo_area": [fi["feature_props"]["fo"]["area"]],
                "fo_intensity": [fi["feature_props"]["fo"]["intensity"]],
                "fo_angle": [fi["feature_axis"]["fo"]["OrientationAngle"]["MajorAxis"]]
            }
            if "intensity_50" in fi["feature_props"]["ob"].keys():
                q = [0,5,10,25,50, 75, 90, 95, 100]
                for k in range(len(q)):
                    dict1["ob_intensity_" + str(q[k])] = [fi["feature_props"]["ob"]["intensity_" + str(q[k])]]
                    dict1["fo_intensity_" + str(q[k])] = [fi["feature_props"]["fo"]["intensity_" + str(q[k])]]

            keys = ["cent_dist", "angle_diff", "area_ratio", "int_area", "bearing",
                    "bdelta", "haus", "medMiss", "medFalseAlarm", "msdMiss", "msdFalseAlarm", "ph", "fom", "minsep"]
            for key in keys:
                dict1[key] = fi["feature_comps"][key]
            df = pd.DataFrame(dict1)
            df_list.append(df)
        if len(df_list)>0:
            df = pd.concat(df_list,axis=0)
            df = df.sort_values(by=["level", "time", "dtime"])
        else:
            df = None
        return  hmfc,df
    except:
        print("将以下检验概要信息转换成DataFrame失败")
        print(features)



def features_tran_miss(features):
    # try:
        time1 = meteva.base.all_type_time_to_time64(features["time"])
        miss_labels = features["unmatched"]['ob']
        df_list = []
        for i in miss_labels:
            fi = features[i]
            dict1 = {
                "level": [0], "time": [time1], "dtime": [features["dtime"]], "id": [i],
                "lon":[fi["feature_props"]["ob"]["centroid"]["x"]],"lat":[fi["feature_props"]["ob"]["centroid"]["y"]],
                "ob_lenghts_MajorAxis":[fi["feature_axis"]["ob"]["lengths"]["MajorAxis"]],
                "ob_lenghts_MinorAxis": [fi["feature_axis"]["ob"]["lengths"]["MinorAxis"]],
                "ob_window_x0":[fi["feature_axis"]["ob"]["window"]["x0"]],
                "ob_window_y0": [fi["feature_axis"]["ob"]["window"]["y0"]],
                "ob_window_x1": [fi["feature_axis"]["ob"]["window"]["x1"]],
                "ob_window_y1": [fi["feature_axis"]["ob"]["window"]["y1"]],
                "ob_centroid_x": [fi["feature_props"]["ob"]["centroid"]["x"]],
                "ob_centroid_y": [fi["feature_props"]["ob"]["centroid"]["y"]],
                "ob_area": [fi["feature_props"]["ob"]["area"]],
                "ob_intensity": [fi["feature_props"]["ob"]["intensity"]],
                "ob_angle": [fi["feature_axis"]["ob"]["OrientationAngle"]["MajorAxis"]],
            }
            if "intensity_50" in fi["feature_props"]["ob"].keys():
                q = [0,5,10,25,50, 75, 90, 95, 100]
                for k in range(len(q)):
                    dict1["ob_intensity_" + str(q[k])] = [fi["feature_props"]["ob"]["intensity_" + str(q[k])]]
            df = pd.DataFrame(dict1)
            df_list.append(df)
        if len(df_list)>0:
            df = pd.concat(df_list,axis=0)
        else:
            df = None
        return  df
    # except:
    #     print("将以下检验概要信息转换成DataFrame失败")
    #     print(features)



def features_tran_false_alarm(features):
    try:
        time1 = meteva.base.all_type_time_to_time64(features["time"])
        false_alarm_labels = features["unmatched"]['fo']
        df_list = []
        for i in false_alarm_labels:
            fi = features[i]
            dict1 = {
                "level": [0], "time": [time1], "dtime": [features["dtime"]], "id": [i],
                "lon":[fi["feature_props"]["fo"]["centroid"]["x"]],"lat":[fi["feature_props"]["fo"]["centroid"]["y"]],
                "fo_lenghts_MajorAxis": [fi["feature_axis"]["fo"]["lengths"]["MajorAxis"]],
                "fo_lenghts_MinorAxis": [fi["feature_axis"]["fo"]["lengths"]["MinorAxis"]],
                "fo_window_x0": [fi["feature_axis"]["fo"]["window"]["x0"]],
                "fo_window_y0": [fi["feature_axis"]["fo"]["window"]["y0"]],
                "fo_window_x1": [fi["feature_axis"]["fo"]["window"]["x1"]],
                "fo_window_y1": [fi["feature_axis"]["fo"]["window"]["y1"]],
                "fo_centroid_x": [fi["feature_props"]["fo"]["centroid"]["x"]],
                "fo_centroid_y": [fi["feature_props"]["fo"]["centroid"]["y"]],
                "fo_area": [fi["feature_props"]["fo"]["area"]],
                "fo_intensity": [fi["feature_props"]["fo"]["intensity"]],
                "fo_angle": [fi["feature_axis"]["fo"]["OrientationAngle"]["MajorAxis"]]
            }
            if "intensity_50" in fi["feature_props"]["fo"].keys():
                q = [0,5,10,25,50, 75, 90, 95, 100]
                for k in range(len(q)):
                    dict1["fo_intensity_" + str(q[k])] = [fi["feature_props"]["fo"]["intensity_" + str(q[k])]]
            df = pd.DataFrame(dict1)
            df_list.append(df)
        if len(df_list)>0:
            df = pd.concat(df_list,axis=0)
        else:
            df = None
        return  df
    except:
        print("将以下检验概要信息转换成DataFrame失败")
        print(features)


def features_list_to_df_old(feature_list):
    hit_list = []
    hmfc_list = []
    for i in range(len(feature_list)):
        features = feature_list[i]
        hmfc,hit = features_tran(features)
        if hmfc is not None and hit is not None:
            hit_list.append(hit)
            hmfc_list.append(hmfc)
    #print(df_list)
    if len(hit_list) > 0:
        df_hit = pd.concat(hit_list,axis=0)
        df_hit = df_hit.reset_index(drop=True)
        hmfc_all = pd.concat(hmfc_list,axis=0)
        hmfc_all = hmfc_all.reset_index(drop=True)
        return hmfc_all,df_hit
    else:
        return None,None



def features_list_to_df(feature_list):
    hit_list = []
    hmfc_list = []
    mis_list = []
    fal_list = []
    for i in range(len(feature_list)):
        features = feature_list[i]
        hmfc,hit = features_tran(features)
        if hmfc is None:
            return None,None,None,None
        mis = features_tran_miss(features)
        fal = features_tran_false_alarm(features)
        hit_list.append(hit)
        hmfc_list.append(hmfc)
        mis_list.append(mis)
        fal_list.append(fal)

    hmfc_all = pd.concat(hmfc_list,axis=0)
    hmfc_all = hmfc_all.reset_index(drop=True)
    df_hit = pd.concat(hit_list,axis=0)
    df_hit = df_hit.reset_index(drop=True)
    df_mis = pd.concat(mis_list,axis=0)
    df_mis = df_mis.reset_index(drop=True)
    df_fal = pd.concat(fal_list,axis=0)
    df_fal = df_fal.reset_index(drop=True)

    return hmfc_all,df_hit,df_fal,df_mis


import meteva
import numpy as np
import datetime
import copy
import time
import pandas as pd
import os



para_example= {
    "begin_time":datetime.datetime.now() - datetime.timedelta(days =7),
    "end_time":datetime.datetime.now(),
    "station_file":r"H:\task\other\202009-veri_objective_method\sta_info.m3",
    "defalut_value":0,
    "hdf_file_name":"last_week_data.h5",
    "interp": meteva.base.interp_gs_nearest,
    "how_fo":"outer",
    "ob_data":{
        "dir_ob": r"Z:\data\surface\jiany_rr\r1\YYMMDDHH.000",
        "hour":None,
        "read_method": meteva.base.io.read_stadata_from_micaps3,
        "read_para": {},
        "reasonable_value": [0, 1000],
        "operation":meteva.base.fun.sum_of_sta,
        "operation_para": {"used_coords": ["time"], "span": 24},
        "time_type": "BT",
    },
    "fo_data":{
        "ECMWF": {
            "dir_fo": r"O:\data\grid\ECMWF_HR\APCP\YYYYMMDD\YYMMDDHH.TTT.nc",
            "hour":[8,20,12],
            "dtime":[0,240,12],
            "read_method": meteva.base.io.read_griddata_from_nc,
            "read_para": {},
            "reasonable_value": [0, 1000],
            "operation": meteva.base.fun.change,
            "operation_para": {"used_coords": "dtime", "delta": 24},
            "time_type": "BT",
            "move_fo_time": 12
        },

        "SCMOC": {
            "dir_fo": r"O:\data\grid\NWFD_SCMOC\RAIN03\YYYYMMDD\YYMMDDHH.TTT.nc",
            "hour": [8, 20,12],
            "dtime":[3,240,3],
            "read_method": meteva.base.io.read_griddata_from_nc,
            "read_para": {},
            "reasonable_value": [0, 1000],
            "operation": meteva.base.fun.sum_of_sta,
            "operation_para": {"used_coords": ["dtime"], "span": 24},
            "time_type": "BT",
            "move_fo_time": 0
        },
    },
    "output_dir":r"H:\test_data\output\mpd\application"
}


def prepare_dataset(para,recover = True):
    '''

    :param para: 根据配置参数从站点和网格数据中读取数据插值到指定站表上，在存储成hdf格式文件，然后从hdf格式文件中读取相应的文件合并成检验要的数据集合文件
    :return:
    '''
    if "dir_ob" in para["ob_data"].keys():

        # 全局参数预处理，站点列表的读取
        station = meteva.base.read_station(para["station_file"])
        station.iloc[:,-1] = para["defalut_value"]
        para["station"] = station
        #全局参数预处理，起止日期的处理
        #day_num = para["day_num"]
        end_time = para["end_time"]
        if end_time is None:
            end_time = datetime.datetime.now()
        #end_date = datetime.datetime(end_time.year, end_time.month, end_time.day, 0, 0) + datetime.timedelta(days=1)
        end_date = datetime.datetime(end_time.year, end_time.month, end_time.day, end_time.hour, 0)

        begin_time = para["begin_time"]
        if begin_time is None:
            begin_time = end_time - datetime.timedelta(days=7)
        begin_date = datetime.datetime(begin_time.year, begin_time.month, begin_time.day, 0, 0)

        hdf_filename = para["hdf_file_name"]
        filename1, type1 = os.path.splitext(hdf_filename)

        para["begin_date"] = begin_date
        para["end_date"] = end_date
        para["day_num"] = int((end_date - begin_date).total_seconds()/3600/24)
        if "hdf_dir" in para["ob_data"].keys():
            hdf_path = para["ob_data"]["hdf_dir"] + "/" + para["hdf_file_name"]
        else:
            hdf_path = para["output_dir"] +"/"+filename1+ "/ob_data/"+hdf_filename

        hdf_file_list = [hdf_path]
        para["ob_data"]["hdf_path"] = hdf_path
        sta_ob = creat_ob_dataset(para)
        operation = para["ob_data"]["operation"]
        operation_para = para["ob_data"]["operation_para"]
        if operation_para is None:
            operation_para = {}
        if operation is not None:
            sta_ob = operation(sta_ob,**operation_para)


        sta_fo_list = []
        models = para["fo_data"].keys()
        for model in models:
            if "hdf_dir" in para["fo_data"][model].keys():
                hdf_path = para["fo_data"][model]["hdf_dir"] + "/" + para["hdf_file_name"]
            else:
                hdf_path = para["output_dir"] + "/" + filename1 + "/fo_" + model + "/" + hdf_filename
            para["fo_data"][model]["hdf_path"] = hdf_path
            hdf_file_list.append(hdf_path)
            sta_fo = creat_fo_dataset(model,para)

            operation = para["fo_data"][model]["operation"]
            operation_para =  para["fo_data"][model]["operation_para"]
            if operation_para is None:
                operation_para = {}
            move_fo_time =  para["fo_data"][model]["move_fo_time"]

            if operation is not None:
                sta_fo = operation(sta_fo, **operation_para)
            if move_fo_time != 0:
                sta_fo = meteva.base.move_fo_time(sta_fo, move_fo_time)
            sta_fo_list.append(sta_fo)
        start = time.time()
        #print(sta_ob)
        #print(sta_fo_list)
        how_fo = "inner"
        if "how_fo" in para.keys():
            how_fo = para["how_fo"]
        sta_all = meteva.base.combine_on_obTime_id(sta_ob,sta_fo_list,how_fo=how_fo)
        if "hdf_dir" in para["ob_data"].keys():
            output_file = para["output_dir"]  + "/" + para["hdf_file_name"]
        else:
            output_file = para["output_dir"] +"/"+filename1+ "/" + para["hdf_file_name"]
        meteva.base.creat_path(output_file)
        if os.path.exists(output_file):
            os.remove(output_file)
        sta_all.to_hdf(output_file, "df")
        print("success combined data to " + output_file)
    else:
        prepare_dataset_without_combining(para,recover = recover)

def prepare_dataset_without_combining(para,recover = True):
    '''

    :param para: 根据配置参数从站点和网格数据中读取数据插值到指定站表上，在存储成hdf格式文件，然后从hdf格式文件中读取相应的文件合并成检验要的数据集合文件
    :return:
    '''

    # 全局参数预处理，站点列表的读取
    station = meteva.base.read_station(para["station_file"])
    station.iloc[:,-1] = para["defalut_value"]
    para["station"] = station

    #全局参数预处理，起止日期的处理
    #day_num = para["day_num"]
    end_time = para["end_time"]
    if end_time is None:
        end_time = datetime.datetime.now()
    end_date = datetime.datetime(end_time.year, end_time.month, end_time.day, 0, 0) + datetime.timedelta(days=1)

    begin_time = para["begin_time"]
    if begin_time is None:
        begin_time = end_time - datetime.timedelta(days=7)
    begin_date = datetime.datetime(begin_time.year, begin_time.month, begin_time.day, 0, 0)

    para["begin_date"] = begin_date
    para["end_date"] = end_date
    para["day_num"] = int((end_date - begin_date).total_seconds()/3600/24)
    hdf_filename = para["hdf_file_name"]
    filename1, type1 = os.path.splitext(hdf_filename)

    elements = para["ob_data"].keys()
    if "dir_ob" in elements:
        if "hdf_dir" in para["ob_data"].keys():
            hdf_path = para["ob_data"]["hdf_dir"] + "/" + para["hdf_file_name"]
        else:
            hdf_path = para["output_dir"] +"/"+filename1+ "/ob_data/"+hdf_filename

        para["ob_data"]["hdf_path"] = hdf_path
        creat_ob_dataset(para)
    else:
        for ele in elements:
            para1 = copy.deepcopy(para)
            if "hdf_dir" in para["ob_data"][ele].keys():
                hdf_path = para["ob_data"][ele]["hdf_dir"] + "/" + para["hdf_file_name"]
            else:
                hdf_path = para["output_dir"]  +"/"+filename1+ "/ob_" +ele+ "/"+hdf_filename

            para1["ob_data"] = para["ob_data"][ele]
            para1["ob_data"]["hdf_path"] = hdf_path
            creat_ob_dataset(para1,ele,recover = recover)

    models = para["fo_data"].keys()
    for model in models:
        if "hdf_dir" in para["fo_data"][model].keys():
            hdf_path = para["fo_data"][model]["hdf_dir"] + "/" + para["hdf_file_name"]
        else:
            hdf_path = para["output_dir"] + "/" + filename1 + "/fo_" + model + "/" + hdf_filename
        para["fo_data"][model]["hdf_path"] = hdf_path
        creat_fo_dataset(model,para)

def creat_fo_dataset(model,para):
    station = para["station"]
    interp = para["interp"]
    end_date = para["end_date"]
    begin_date = para["begin_date"]
    day_num = para["day_num"] + 1
    para_model = para["fo_data"][model]
    hdf_path = para_model["hdf_path"]
    dir_fo  =para_model["dir_fo"]
    read_method = para_model["read_method"]
    read_para =para_model["read_para"]
    reasonable_value = para_model["reasonable_value"]
    if read_para is None:
        read_para = {}
    data0 = None
    if os.path.exists(hdf_path):
        data0 = pd.read_hdf(hdf_path, "df")

    hours = None
    if para_model["hour"] is not None:
        hours = np.arange(para_model["hour"][0],para_model["hour"][1]+1,para_model["hour"][2]).tolist()


    dtimes = None
    if para_model["dtime"] is not None:
        dtimes = np.arange(para_model["dtime"][0],para_model["dtime"][1]+1,para_model["dtime"][2]).tolist()

    sta_list = [] #用于收集所有数据
    exist_dtimes = {}

    gds_file_list = []
    is_gds = False
    if dir_fo is not None:
        if dir_fo.find("mdfs:") >= 0:
            dir1, gds_filename = os.path.split(dir_fo)
            dir1 = dir1.replace(">", "")
            gds_file_list = meteva.base.path_tools.get_gds_path_list_in_one_dir(dir1)
            is_gds = True

    if data0 is None:
        if hours is None:
            hours = np.arange(0, 24, 1).tolist()
        if dtimes is None:
            dtimes = np.arange(0, 721, 1).tolist()
    else:
        data_left = meteva.base.sele_by_para(data0, time_range=[begin_date, end_date])
        data_name0 = meteva.base.get_stadata_names(data_left)
        if len(data_name0) == 2:
            data_name1 = ["u_"+model , "v_"+model]
        else:
            data_name1 = [model]
        meteva.base.set_stadata_names(data_left, data_name1)

        #meteva.base.set_stadata_names(data_left,model)
        sta_list.append(data_left)
        #id0 = station["id"].values[0]
        #data_id0 = meteva.base.sele_by_para(data0, id=id0)
        #print(data_id0)

        grouped_dict = dict(list(data_left.groupby("time")))
        keys = grouped_dict.keys()
        exist_time_list = []
        for key in keys:
            time1 = meteva.base.time_tools.all_type_time_to_datetime(key)
            exist_time_list.append(time1)
            ehours = grouped_dict[time1].loc[:, "dtime"].values.tolist()
            exist_dtimes[time1] = ehours
            #valid_group_list_list.append([key])
            #sta_ob_and_fos_list.append(grouped_dict[key])



        if hours is None:
            hours = []
            for time1 in exist_time_list:
                hours.append(time1.hour)
            hours = list(set(hours))
            hours.sort()
        if dtimes is None:
            dtimes = data_left.loc[:, "dtime"].values.tolist()
            dtimes = list(set(dtimes))
            dtimes.sort()
        if len(hours) == 0:
            hours = np.arange(0,24,1).tolist()
        if len(dtimes) == 0:
            dtimes = np.arange(0, 721, 1).tolist()


    #print(exist_dtimes)

    for dd in range(day_num):
        dati_s = end_date  - datetime.timedelta(days=dd)
        dati_s = datetime.datetime(dati_s.year,dati_s.month,dati_s.day,0,0)
        for hh in range(len(hours)):
            hour = hours[hh]
            time1 =dati_s  + datetime.timedelta(hours=hour)

            if time1 > end_date or time1< begin_date:continue
            if para_model["time_type"] == "BT":
                file_time = time1
            else:
                file_time = time1 - datetime.timedelta(hours = 8)


            for dt in dtimes:
                #data_exist = False
                if time1 in exist_dtimes.keys():
                    exist_dtime = exist_dtimes[time1]
                    if dt in exist_dtime:
                        #data_exist = True
                        continue
                #if data_exist:continue
                if dir_fo is None:
                    dat = read_method(**read_para,time = file_time,dtime = dt)
                    if dat is not None:
                        if not isinstance(dat, pd.DataFrame):
                            dat = interp(dat, station)
                        else:
                            dat = meteva.base.put_stadata_on_station(dat, station)

                        if reasonable_value is not None:
                            dat = meteva.base.sele_by_para(dat, value=reasonable_value)

                        meteva.base.set_stadata_coords(dat, time=time1, dtime=dt)
                        data_name0 = meteva.base.get_stadata_names(dat)
                        if len(data_name0) == 2:
                            data_name1 = ["u_"+model , "v_"+model]
                        else:
                            data_name1 = [model]
                        meteva.base.set_stadata_names(dat, data_name1)
                        sta_list.append(dat)
                        print("success read data from " + str(read_para)+ str(file_time)+"."+str(dt))
                else:
                    file_exit = False
                    if is_gds:
                        path = meteva.base.get_path(dir_fo, file_time,dt)
                        if path in gds_file_list:
                            file_exit = True
                    else:
                        path = meteva.base.get_path(dir_fo, file_time,dt)
                        if os.path.exists(path) or path is None:
                            file_exit = True

                    if file_exit:
                    #if os.path.exists(path) or path is None:
                        try:
                            dat = read_method(path,**read_para)
                            if dat is not None:
                                if not isinstance(dat, pd.DataFrame):
                                    dat = interp(dat, station)
                                else:
                                    dat = meteva.base.put_stadata_on_station(dat,station)

                                if reasonable_value is not None:
                                    dat = meteva.base.sele_by_para(dat, value=reasonable_value)

                                meteva.base.set_stadata_coords(dat,time = time1,dtime = dt)
                                data_name0 = meteva.base.get_stadata_names(dat)
                                if len(data_name0) ==2:
                                    data_name1 = ["u_"+model , "v_"+model]
                                else:
                                    data_name1 = [model]
                                meteva.base.set_stadata_names(dat,data_name1)
                                sta_list.append(dat)
                                print("success read data from " + path)
                            else:
                                print("fail read data from " + path)
                        except:
                            print("fail read data from " + path)
                    else:
                        print(path +" does not exist")

    if(len(sta_list) == 0):
        print("there is not file data in " + dir_fo)
        return
    sta_all = pd.concat(sta_list, axis=0)
    if "level" not in read_para.keys():
        meteva.base.set_stadata_coords(sta_all, level=0)

    meteva.base.creat_path(hdf_path)
    if os.path.exists(hdf_path):
        os.remove(hdf_path)
    sta_all.to_hdf(hdf_path, "df")
    print(hdf_path)
    return sta_all

def creat_ob_dataset(para,ele = "ob",recover = True):
    station = para["station"]
    data_name =ele
    day_num = para["day_num"] + 1
    end_date = para["end_date"]
    begin_date = para["begin_date"]

    hdf_path = para["ob_data"]["hdf_path"]
    dir_ob =para["ob_data"]["dir_ob"]
    read_method = para["ob_data"]["read_method"]
    read_para = para["ob_data"]["read_para"]
    if read_para is None:
        read_para = {}
    reasonable_value = para["ob_data"]["reasonable_value"]

    hours = None
    if para["ob_data"]["hour"] is not None:
        hours = np.arange(para["ob_data"]["hour"][0],para["ob_data"]["hour"][1]+1,para["ob_data"]["hour"][2]).tolist()



    exist_time_list = []
    sta_list = []
    data0 = None
    if os.path.exists(hdf_path):
        data0 = pd.read_hdf(hdf_path, "df")
        if not recover:
            return data0
    if data0 is None:
        if hours is None:
            hours = np.arange(0, 24, 1).tolist()
    else:
        data_left = meteva.base.sele_by_para(data0, time_range=[begin_date, end_date])
        data_name0 = meteva.base.get_stadata_names(data_left)
        if len(data_name0) == 1:
            meteva.base.set_stadata_names(data_left, data_name)
        sta_list.append(data_left)
        #id0 = station["id"].values[0]
        #data_id0 = meteva.base.sele_by_para(data0, id=id0)
        times = data_left.loc[:, "time"].values.tolist()
        times = list(set(times))
        times.sort()

        for i in range(len(times)):
            time1 = meteva.base.time_tools.all_type_time_to_datetime(times[i])
            exist_time_list.append(time1)
        if hours is None:
            hours = []
            for time1 in exist_time_list:
                hours.append(time1.hour)
            hours = list(set(hours))
            hours.sort()
        if len(hours) == 0:
            hours = np.arange(0,24,1).tolist()


    gds_file_list = []
    is_gds = False
    if dir_ob is not None:
        if dir_ob.find("mdfs:") >= 0:
            dir1, gds_filename = os.path.split(dir_ob)
            gds_file_list = meteva.base.path_tools.get_gds_path_list_in_one_dir(dir1)
            if(len(gds_file_list) == 0):
                if dir1.find("HH")>=0:
                    for hh in hours:
                        dir2 = dir1.replace("HH","%02d"%hh)
                        gds_file_list2 = meteva.base.path_tools.get_gds_path_list_in_one_dir(dir2)
                        gds_file_list.extend(gds_file_list2)
            is_gds = True

    for dd in range(day_num):
        dati_s = end_date  - datetime.timedelta(days=dd)
        dati_s = datetime.datetime(dati_s.year,dati_s.month,dati_s.day,0,0)

        for hh in range(len(hours)):
            hour = hours[hh]
            #time1 = end_date - datetime.timedelta(days=dd) + datetime.timedelta(hours=hour)
            time1 = dati_s + datetime.timedelta(hours=hour)
            if time1 > end_date or time1 < begin_date: continue
            if time1 in exist_time_list:
                continue
            if para["ob_data"]["time_type"] == "BT":
                file_time = time1
            else:
                file_time = time1 - datetime.timedelta(hours = 8)

            if dir_ob is None:
                dat = read_method(**read_para, time=file_time)
                if dat is not None:
                    if not isinstance(dat, pd.DataFrame):
                        interp = para["interp"]
                        dat = interp(dat, station)
                    else:
                        dat = meteva.base.fun.comp.put_stadata_on_station(dat, station)
                    if reasonable_value is not None:
                        dat = meteva.base.sele_by_para(dat, value=reasonable_value)
                    data_name0 = meteva.base.get_stadata_names(dat)
                    if len(data_name0) == 1:
                        meteva.base.set_stadata_names(dat, data_name)
                    meteva.base.set_stadata_coords(dat, time=time1)
                    sta_list.append(dat)
                    print("success read data from " + str(read_para) + str(file_time))
            else:
                file_exit = False
                path = meteva.base.get_path(dir_ob, file_time)
                if is_gds:
                    if path in gds_file_list:
                        file_exit = True
                else:
                    if os.path.exists(path) or path is None:
                        file_exit = True
                if file_exit:
                    try:
                        dat = read_method(path,**read_para)
                        if dat is not None:
                            dat = meteva.base.fun.comp.put_stadata_on_station(dat,station)
                            if not isinstance(dat,pd.DataFrame):
                                interp = para["interp"]
                                dat = interp(dat,station)
                            if reasonable_value is not None:
                                dat = meteva.base.sele_by_para(dat,value=reasonable_value)
                            data_name0 = meteva.base.get_stadata_names(dat)
                            if len(data_name0) == 1:
                                meteva.base.set_stadata_names(dat,data_name)
                            meteva.base.set_stadata_coords(dat,time = time1)
                            sta_list.append(dat)
                            print("success read data from "+ path)
                        else:
                            print("fail read data from " + path)
                    except:
                        print("fail read data from " + path)
                else:
                    print(path +  "does not exist")
    if(len(sta_list)==0):return None
    sta_all = pd.concat(sta_list, axis=0)
    if "level" not in read_para.keys():
        meteva.base.set_stadata_coords(sta_all, level=0)

    meteva.base.creat_path(hdf_path)
    if os.path.exists(hdf_path):
        os.remove(hdf_path)
    sta_all.to_hdf(hdf_path, "df")
    #print(hdf_path)

    return sta_all


def rename_hdf_file(old_para,new_para):
    '''
    更改数据收集参数中的hdf文件名称时，涉及到多个中间文件。当需要更改文件名称时，通过该程序自动的完成所有文件的重命名
    :param old_hdf_filename:  更改前的文件名称
    :param para:  包含更改后文件名的运行参数
    :return:  不返还值
    '''
    import shutil
    # 更改观测的hdf文件
    old_hdf_filename = old_para["hdf_file_name"]
    new_hdf_filename = new_para["hdf_file_name"]
    filename1,type1 = os.path.splitext(new_hdf_filename)
    if "dir_ob" in old_para["ob_data"].keys():
        filename_old = old_para["ob_data"]["hdf_dir"] + "/"+old_hdf_filename
        filename_new = new_para["output_dir"] +"/"+filename1+ "/ob_data/"+new_hdf_filename
        meteva.base.creat_path(filename_new)
        try:
            shutil.move(filename_old,filename_new)
        except:
            print(filename_old+"移动失败")
    else:
        elements = old_para["ob_data"].keys()
        for ele in elements:
            filename_old = old_para["ob_data"][ele]["hdf_dir"] + "/" + old_hdf_filename
            filename_new = new_para["output_dir"]  +"/"+filename1+ "/ob_" +ele+ "/"+new_hdf_filename
            meteva.base.creat_path(filename_new)
            shutil.move(filename_old, filename_new)
    # 重命名预报的hdf文件
    elements = old_para["fo_data"].keys()
    for ele in elements:
        filename_old = old_para["fo_data"][ele]["hdf_dir"] + "/" + old_hdf_filename
        filename_new = new_para["output_dir"]  +"/"+filename1+ "/fo_" +ele+ "/"+new_hdf_filename
        meteva.base.creat_path(filename_new)
        shutil.move(filename_old, filename_new)
    return

if __name__ == "__main__":
    pass
    #prepare_dataset(para_example)



#import statsmodels.api as sm
import numpy as np

# 回归函数，已通过sma.py模块中的函数代替，暂留备份
#
# def regress2(_x, _y, _method_type_1 = "ordinary least square",
#              _method_type_2 = "reduced major axis",
#              _weight_x = [], _weight_y = [], _need_intercept = True):
#     # Regression Type II based on statsmodels
#     # Type II regressions are recommended if there is variability on both x and y
#     # It's computing the linear regression type I for (x,y) and (y,x)
#     # and then average relationship with one of the type II methods
#     #
#     # INPUT:
#     #   _x <np.array>
#     #   _y <np.array>
#     #   _method_type_1 <str> method to use for regression type I:
#     #     ordinary least square or OLS <default>
#     #     weighted least square or WLS
#     #     robust linear model or RLM
#     #   _method_type_2 <str> method to use for regression type II:
#     #     major axis
#     #     reduced major axis <default> (also known as geometric mean)
#     #     arithmetic mean
#     #   _need_intercept <bool>
#     #     True <default> add a constant to relation (y = a x + b)
#     #     False force relation by 0 (y = a x)
#     #   _weight_x <np.array> containing the weigth of x
#     #   _weigth_y <np.array> containing the weigth of y
#     #
#     # OUTPUT:
#     #   slope
#     #   intercept
#     #   r
#     #   std_slope
#     #   std_intercept
#     #   predict
#     #
#     # REQUIRE:
#     #   numpy
#     #   statsmodels
#     #
#     # The code is based on the matlab function of MBARI.
#     # AUTHOR: Nils Haentjens
#     # REFERENCE: https://www.mbari.org/products/research-software/matlab-scripts-linear-regressions/
#
#     # Check input
#     if _method_type_2 != "reduced major axis" and _method_type_1 != "ordinary least square":
#         raise ValueError("'" + _method_type_2 + "' only supports '" + _method_type_1 + "' method as type 1.")
#
#     # Set x, y depending on intercept requirement
#     if _need_intercept:
#         x_intercept = sm.add_constant(_x)
#         y_intercept = sm.add_constant(_y)
#
#     # Compute Regression Type I (if type II requires it)
#     if (_method_type_2 == "reduced major axis" or
#         _method_type_2 == "geometric mean"):
#         if _method_type_1 == "OLS" or _method_type_1 == "ordinary least square":
#             if _need_intercept:
#                 [intercept_a, slope_a] = sm.OLS(_y, x_intercept).fit().params
#                 [intercept_b, slope_b] = sm.OLS(_x, y_intercept).fit().params
#             else:
#                 slope_a = sm.OLS(_y, _x).fit().params
#                 slope_b = sm.OLS(_x, _y).fit().params
#         elif _method_type_1 == "WLS" or _method_type_1 == "weighted least square":
#             if _need_intercept:
#                 [intercept_a, slope_a] = sm.WLS(
#                     _y, x_intercept, weights=1. / _weight_y).fit().params
#                 [intercept_b, slope_b] = sm.WLS(
#                     _x, y_intercept, weights=1. / _weight_x).fit().params
#             else:
#                 slope_a = sm.WLS(_y, _x, weights=1. / _weight_y).fit().params
#                 slope_b = sm.WLS(_x, _y, weights=1. / _weight_x).fit().params
#         elif _method_type_1 == "RLM" or _method_type_1 == "robust linear model":
#             if _need_intercept:
#                 [intercept_a, slope_a] = sm.RLM(_y, x_intercept).fit().params
#                 [intercept_b, slope_b] = sm.RLM(_x, y_intercept).fit().params
#             else:
#                 slope_a = sm.RLM(_y, _x).fit().params
#                 slope_b = sm.RLM(_x, _y).fit().params
#         else:
#             raise ValueError("Invalid literal for _method_type_1: " + _method_type_1)
#
#     # Compute Regression Type II
#     if (_method_type_2 == "reduced major axis" or
#         _method_type_2 == "geometric mean"):
#         # Transpose coefficients
#         if _need_intercept:
#             intercept_b = -intercept_b / slope_b
#         slope_b = 1 / slope_b
#         # Check if correlated in same direction
#         #if np.sign(slope_a) != np.sign(slope_b):
#         #    raise RuntimeError('Type I regressions of opposite sign.')
#         # Compute Reduced Major Axis Slope
#         slope = np.sign(slope_a) * np.sqrt(slope_a * slope_b)
#         if _need_intercept:
#             # Compute Intercept (use mean for least square)
#             if _method_type_1 == "OLS" or _method_type_1 == "ordinary least square":
#                 intercept = np.mean(_y) - slope * np.mean(_x)
#             else:
#                 intercept = np.median(_y) - slope * np.median(_x)
#         else:
#             intercept = 0
#         # Compute r
#         r = np.sign(slope_a) * np.sqrt(slope_a / slope_b)
#         # Compute predicted values
#         predict = slope * _x + intercept
#         # Compute standard deviation of the slope and the intercept
#         n = len(_x)
#         diff = _y - predict
#         Sx2 = np.sum(np.multiply(_x, _x))
#         den = n * Sx2 - np.sum(_x) ** 2
#         s2 = np.sum(np.multiply(diff, diff)) / (n - 2)
#         std_slope = np.sqrt(n * s2 / den)
#         if _need_intercept:
#             std_intercept = np.sqrt(Sx2 * s2 / den)
#         else:
#             std_intercept = 0
#     elif (_method_type_2 == "Pearson's major axis" or
#           _method_type_2 == "major axis"):
#         if not _need_intercept:
#             raise ValueError("Invalid value for _need_intercept: " + str(_need_intercept))
#         xm = np.mean(_x)
#         ym = np.mean(_y)
#         xp = _x - xm
#         yp = _y - ym
#         sumx2 = np.sum(np.multiply(xp, xp))
#         sumy2 = np.sum(np.multiply(yp, yp))
#         sumxy = np.sum(np.multiply(xp, yp))
#         slope = ((sumy2 - sumx2 + np.sqrt((sumy2 - sumx2)**2 + 4 * sumxy**2)) /
#                  (2 * sumxy))
#         intercept = ym - slope * xm
#         # Compute r
#         r = sumxy / np.sqrt(sumx2 * sumy2)
#         # Compute standard deviation of the slope and the intercept
#         n = len(_x)
#         std_slope = (slope / r) * np.sqrt((1 - r ** 2) / n)
#         sigx = np.sqrt(sumx2 / (n - 1))
#         sigy = np.sqrt(sumy2 / (n - 1))
#         std_i1 = (sigy - sigx * slope) ** 2
#         std_i2 = (2 * sigx * sigy) + ((xm ** 2 * slope * (1 + r)) / r ** 2)
#         std_intercept = np.sqrt((std_i1 + ((1 - r) * slope * std_i2)) / n)
#         # Compute predicted values
#         predict = slope * _x + intercept
#     elif _method_type_2 == "arithmetic mean":
#         if not _need_intercept:
#             raise ValueError("Invalid value for _need_intercept: " + str(_need_intercept))
#         n = len(_x)
#         sg = np.floor(n / 2)
#         # Sort x and y in order of x
#         sorted_index = sorted(range(len(_x)), key=lambda i: _x[i])
#         x_w = np.array([_x[i] for i in sorted_index])
#         y_w = np.array([_y[i] for i in sorted_index])
#         x1 = x_w[1:sg + 1]
#         x2 = x_w[sg:n]
#         y1 = y_w[1:sg + 1]
#         y2 = y_w[sg:n]
#         x1m = np.mean(x1)
#         x2m = np.mean(x2)
#         y1m = np.mean(y1)
#         y2m = np.mean(y2)
#         xm = (x1m + x2m) / 2
#         ym = (y1m + y2m) / 2
#         slope = (x2m - x1m) / (y2m - y1m)
#         intercept = ym - xm * slope
#         # r (to verify)
#         r = []
#         # Compute predicted values
#         predict = slope * _x + intercept
#         # Compute standard deviation of the slope and the intercept
#         std_slope = []
#         std_intercept = []
#
#     # Return all that
#     return {"slope": float(slope), "intercept": intercept, "r": r,
#             "std_slope": std_slope, "std_intercept": std_intercept,
#             "predict": predict}






def get_encoding_of_file(filename,read_rows = 0):
    encoding = None
    strs = ""
    try:
        encoding = "GBK"
        file = open(filename, encoding="GBK")
        if read_rows == 0:
            strs = file.read()
        else:
            if read_rows == 1:
                strs = file.readline()
            else:
                strs = []
                for i in range(read_rows):
                    strs.append(file.readline())
        file.close()

    except:
        try:
            encoding = "UTF-8"
            file = open(filename, encoding="UTF-8")
            if read_rows == 0:
                strs = file.read()
            else:
                if read_rows == 1:
                    strs = file.readline()
                else:
                    strs = []
                    for i in range(read_rows):
                        strs.append(file.readline())
            file.close()
        except:
            try:
                encoding = "GB2312"
                file = open(filename, encoding="GB2312")
                if read_rows == 0:
                    strs = file.read()
                else:
                    if read_rows == 1:
                        strs = file.readline()
                    else:
                        strs = []
                        for i in range(read_rows):
                            strs.append(file.readline())
                file.close()
            except:
                print(filename + "文件编码不是GBK、GB2312或者UTF-8格式，程序暂时不能识别")
                return None
    return encoding,strs


#!/usr/bin/python
# -*- coding: utf-8 -*-
#Filename: CMADaasAccess.py

'''
Created on Sep 15, 2020

@author: anduin
'''
import meteva
import datetime
import json
from .loglib import *
from .httpconn import HttpConn
from .SignGenUtil import SignGenUtil
import numpy as np

class CMADaasAccess():
    """description of class 
    cmadaas data access
    """    
    @staticmethod
    def get_times(dts):       
        try:
            if type(dts) is list:
                return ','.join(dt.strftime('%Y%m%d%H%M%S') for dt in dts)
            else:
                return dts.strftime('%Y%m%d%H%M%S')
        except Exception as data:
            #print(str(data))           
            LogLib.error('CMADaasAccess get_times except %s %s' % (str(dts), str(data)))
            raise data
            #return None
    
    @staticmethod
    def set_times_in_params(dts, params, deep_copy=True):
        try:
            rst = params
            if deep_copy:
                import copy
                rst = copy.deepcopy(params)

            rst['times'] = CMADaasAccess.get_times(dts)

            return rst
        except Exception as data:
            #print(str(data))
            LogLib.error('CMADaasAccess set_times_in_params except %s %s' % (str(dts), str(data)))
            raise data
            #return None
    
    @staticmethod
    def get_timerange(dts):
        try:
            rst = ''
            if type(dts) is list:
                rst = ','.join(dt.strftime('%Y%m%d%H%M%S') for dt in dts)
            else:
                rst = dts.strftime('%Y%m%d%H%M%S')

            return '[' + rst + ']'
        except Exception as data:
            #print(str(data)) 
            LogLib.error('CMADaasAccess get_timerange except %s %s' % (str(dts), str(data)))
            raise data
            #return None
    
    @staticmethod
    def set_timerange_in_params(dts, params, deep_copy=True):
        try:
            rst = params
            if deep_copy:
                import copy
                rst = copy.deepcopy(params)

            rst['timeRange'] = CMADaasAccess.get_timerange(dts)

            return rst
        except Exception as data:
            #print(str(data))
            LogLib.error('CMADaasAccess set_timerange_in_params except %s %s' % (str(dts), str(data)))
            raise data
            #return None
    
    @staticmethod
    def get_query(params):
        '''
        生成url中query部分
        '''
        try:
            querystr = ''
            for key,item in params.items():
                querystr += key + "=" + str(item) + "&"

            if(querystr):
                querystr = querystr[:-1]
            
            return querystr
        except Exception as data:
            #print(str(data))
            LogLib.error('CMADaasAccess get_query except %s %s' % (str(params), str(data)))
            raise data
            #return None
            
    @staticmethod
    def get_url(url, qparams):
        '''
        生成url
        '''
        #import copy
        try:
            #signParams = copy.deepcopy(qparams)
            signParams = qparams
            signString = SignGenUtil.getSign(signParams)
            if signString is None:
                LogLib.error('CMADaasAccess getUrl getSign error %s %s' % (url, str(qparams)))
                return None

            signParams['sign'] = signString
            signParams.pop('pwd')

            
            querystr = CMADaasAccess.get_query(signParams)
            
            return url + querystr
        except Exception as data:
            #print(str(data))  
            LogLib.error('CMADaasAccess getUrl except %s %s %s' % (url, str(qparams), str(data)))
            raise data
            #return None
            
    @staticmethod
    def get_default_headers():
        return { 'Accept' : 'application/json, text/plain, */*',
                'Accept-Encoding' : 'gzip, deflate, br',
                'Accept-Language' : 'zh-CN,zh;q=0.9',
                'User-Agent' : 'Mozilla/5.0 (X11; Linux x86_64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/75.0.3770.100 Safari/537.36' }
            
    #only json is supported
    #从cmadaas获得json格式数据，使用json库解析，构建一个dataframe返回。
    @staticmethod
    def read_data_from_cmadaas(fullurl):
        try:
            LogLib.info('CMADaasAccess read_data_from_cmadaas start %s' % (str(fullurl)))
            headers = CMADaasAccess.get_default_headers()
            host, url, ishttp = HttpConn.parseUrl(fullurl)
            rststatus, rstcontent, rstheaders = HttpConn.getWebPage(host, url, headers=headers, ishttp=ishttp)
            if rststatus != 200:
                LogLib.error('CMADaasAccess read_data_from_cmadaas net error %d %s' % (rststatus, str(fullurl)))
                return None

            jsonobj = json.loads(rstcontent)
            if jsonobj is None:
                LogLib.error('CMADaasAccess read_data_from_cmadaas json loads error %s %s' % (rstcontent, str(fullurl)))
                return None

            if 'returnCode' not in jsonobj:
                LogLib.error('CMADaasAccess read_data_from_cmadaas json no returnCode %s %s' % (rstcontent, str(fullurl)))
                return None

            if jsonobj['returnCode'] != '0':
                if jsonobj['returnCode'] == '-1':
                    LogLib.info('CMADaasAccess read_data_from_cmadaas json no data %s %s' % (rstcontent, str(fullurl)))
                    return []
                else:
                    LogLib.error('CMADaasAccess read_data_from_cmadaas json no returnCode or returnCode != 0 %s %s' % (rstcontent, str(fullurl)))
                    return None

            return jsonobj['DS']
        except Exception as data:
            LogLib.error('CMADaasAccess read_data_from_cmadaas except %s %s' % (str(fullurl), str(data)))
            return None
    
    @staticmethod
    def read_griddata_from_cmadaas(fullurl,time=None,dtime=None):
        try:
            LogLib.info('CMADaasAccess read_data_from_cmadaas start %s' % (str(fullurl)))
            headers = CMADaasAccess.get_default_headers()
            host, url, ishttp = HttpConn.parseUrl(fullurl)
            rststatus, rstcontent, rstheaders = HttpConn.getWebPage(host, url, headers=headers, ishttp=ishttp)
            if rststatus != 200:
                LogLib.error('CMADaasAccess read_data_from_cmadaas net error %d %s' % (rststatus, str(fullurl)))
                return None

            jsonobj = json.loads(rstcontent)
            if jsonobj is None:
                LogLib.error('CMADaasAccess read_data_from_cmadaas json loads error %s %s' % (rstcontent, str(fullurl)))
                return None

            if 'returnCode' not in jsonobj:
                LogLib.error('CMADaasAccess read_data_from_cmadaas json no returnCode %s %s' % (rstcontent, str(fullurl)))
                return None
            
            griddata = CMADaasAccess.griddata_from_cmadaas_json(jsonobj,time=time, dtime=dtime)
            return griddata
        except Exception as data:
            LogLib.error('CMADaasAccess read_data_from_cmadaas except %s %s' % (str(fullurl), str(data)))
            return None


    @staticmethod
    def griddata_from_cmadaas_json(json,time=None,dtime=None):
        try:
            lonS = round(json['startLon'],4)
            # lonE = round(json['endLon'],3)
            latS = round(json['startLat'],4)
            # latE = round(json['endLat'],3)
            dlon = round(json['lonStep'],4)
            dlat = round(json['latStep'],4)
            nlon = int(json['lonCount'])
            nlat = int(json['latCount'])
            lonE = lonS + (nlon-1)*dlon
            latE = latS + (nlat-1)*dlat
            #print(lonS,dlon,nlon,lonE)
            data = np.array(json['DS'])
            
            #data
            # print(data.shape)

            if time is not None:
                if not isinstance(time,list): 
                    time = [time]#确保time为列表
            if dtime is not None:
                if not isinstance(dtime,list): 
                    dtime = [dtime]#确保dtime为列表
            # print(time,dtime)
            grid_info = meteva.base.grid([lonS,lonE,dlon],[latS,latE,dlat]
                                 ,gtime = time
                                 ,dtime_list=dtime)
            #print(grid_info)
            grid = meteva.base.grid_data(grid_info,data)
        except Exception as data:
            print(json)
            return None
        return grid
            
                
    
    @staticmethod
    def dataframe_from_dict_json(j_dict,default=np.nan):
        import pandas as pd
        try:
            LogLib.info('CMADaasAccess Convert dataFrame from JSON_dict {0}'.format(datetime.datetime.now().strftime("%Y%m%d%H%M")))
            pddata = pd.DataFrame.from_dict(j_dict)
            ## 更改dataframe
            sta = meteva.base.sta_data(pddata, columns=['time','id','lat','lon','data0'])
            sta.level = 0
            sta.dtime=0
            sta.time = pd.to_datetime(sta.time)
            sta.level = sta.level.astype(np.int16)
            sta.id = sta.id.astype(np.int32)
            sta.dtime = sta.dtime.astype(np.int16)
            sta.lon = sta.lon.astype(np.float32)
            sta.lat = sta.lat.astype(np.float32)
            try:
                sta.data0 = sta.data0.astype(np.float32)
                sta.data0[sta.data0>=990000]=default
            except:
                pass
            return(sta)            
        except Exception as data:
            LogLib.error('CMADaasAccess Convert dataFrame from JSON_dict {0}'.format(datetime.datetime.now().strftime("%Y%m%d%H%M")))
            return(None)
            
        
        
    #从cmadaas获得文件数据。
    @staticmethod
    def read_file_from_cmadaas(fullurl):
        try:
            LogLib.info('CMADaasAccess read_file_from_cmadaas start %s' % (str(fullurl)))
            headers = CMADaasAccess.get_default_headers()
            host, url, ishttp = HttpConn.parseUrl(fullurl)
            rststatus, rstcontent, rstheaders = HttpConn.getWebPage(host, url, headers=headers, ishttp=ishttp)
            if rststatus != 200:
                LogLib.error('CMADaasAccess read_file_from_cmadaas net error %d %s' % (rststatus, str(fullurl)))
                return None
            return (rstcontent, rstheaders)
        except Exception as data:
            LogLib.error('CMADaasAccess read_file_from_cmadaas except %s %s' % (str(fullurl), str(data)))
            return None
        
    @staticmethod    
    def write_data_to_file(save_path,grd,creat_dir=True):
        import os
        try:
            LogLib.info('FixWriteData write_data_to_file start %s' % (save_path))            
            if creat_dir:
                path,file = os.path.split(save_path)
                if not os.path.exists(path):
                   os.makedirs(path)
            with open(save_path, 'wb') as f:
                f.write(grd)
            LogLib.info('FixWriteData write_data_to_file over %s' % (save_path))
            return True
        except Exception as data:
            LogLib.error('FixWriteData write_data_to_file except %s' % (str(save_path)))
            raise data
            
    ## 生成datetime类
    def get_datetime_from_str(dt='202011110200'):#dt可以为任意长度字符串，也可为datetime类
        if type(dt) == str:
            #用户输入2019041910十位字符，后面补全加0000，为14位统一处理
            if len(dt) == 4:
                num1=dt + "0101000000"
            elif len(dt) == 6:
                num1=dt + "01000000"
            elif len(dt) == 8:
                num1=dt + "000000"
            elif len(dt) == 10:
                num1=dt + "0000"
            elif len(dt) == 12:
                num1=dt + "00"
            elif len(dt) == 14:
                num1=dt
            else:
                print("输入日期有误，请检查！")
            #统一将日期变为datetime类型
            time = datetime.datetime.strptime(num1, '%Y%m%d%H%M%S')
        else:
            time = dt
        return(time)
    

    ## 根据参数字典，生成REST格式url。            
    @staticmethod
    def combine_url_from_para(qprams, time=datetime.datetime(2020,11,27,0,0), url=None,
                              SN_id="NMIC_MUSIC_CMADAAS", dataFormat='json', userId=None, pwd=None,
                                     defaultNan=9999, show_url=False, time_name='times'):
        """
        dt： 可以为字符串str，如“202011110200”；也可以为datetime类时间数据
        userID: 业务账号，必须填
        pwd： 业务账户密码，必须填
        dataFormat: 返回数据格式，默认为json
        show_url:  是否完整显示url，默认为不显示
        url: cmadaas的url前面固定部分，默认'http://10.40.17.54/music-ws/api?'
        SN_id: cmadaas服务节点代码，默认'NMIC_MUSIC_CMADAAS'
        time_name:时间参数的字典名(time或者times)
        """
        import copy
        params = copy.copy(qprams)
        LogLib.info('OBS Download from CmaDaas: {0}'.format(datetime.datetime.now().strftime("%Y%m%d%H%M"))) 
            
        ## 账户处理
        if userId is None:
            userId = meteva.base.cmadaas_set[2]
            LogLib.warning('NO userID input,Using default USER: {0}'.format(datetime.datetime.now().strftime("%Y%m%d%H%M")))
        if pwd is None:
            pwd = meteva.base.cmadaas_set[3]
            LogLib.warning('NO Password input,Using default pwd: {0}'.format(datetime.datetime.now().strftime("%Y%m%d%H%M")))
        ## 添加字典变量要素   
        if "timeRange" in params:
            params['timeRange'] = (str(params['timeRange']).replace(' ',''))
        if "userId" not in params:
            params["userId"] = userId
#            params = params.pop("userId")
        if "pwd" not in params:
            params["pwd"] = pwd
        if not (("times" in params) or ("time" in params) or ("timeRange" in params)):
            params[time_name] = CMADaasAccess.get_times(time)
        if "serviceNodeId" not in params: params["serviceNodeId"] = SN_id
        if "timestamp" not in params : params["timestamp"]= str(int(datetime.datetime.now().timestamp()*1000))  
        if "nonce" not in params : params["nonce"] = '3696663f-d202-4570-b39d-16306f419575'
        if "dataFormat" not in params : params['dataFormat'] = dataFormat
        ## 生成url
        
        #print(params['timeRange'])
        if url is None:
            url = "http://"+meteva.base.cmadaas_set[0]+"/music-ws/api?"

        urlstr = CMADaasAccess.get_url(url, params)
        if show_url is True : print(urlstr)
        ## 返回url
        return(urlstr)
        
        
            
    
    @staticmethod            
    def get_obs_micaps3_from_cmadaas(qparams, time="202011110200",save_path=None, filename=None, defaultNan=9999,show = False,
                                     **kargs):
        """
        函数返回：meteva形式的pandas.DataFrame格式；如果申明sava_path变量，可直接输出Micaps3文件
        ###
        qparams: 包含数据、列表、下载要素的字典容器
        time： 可以为字符串str，如“202011110200”；也可以为datetime类时间数据
        userID: 业务账号，必须填
        pwd： 业务账户密码，必须填
        save_path: 存储文件目录，若不填写则不输出
        filename:  存储文件名,可以使用通配符(如YYYYMMDDHHFF.TTT即表示“年(四位)月(两位)日(两位)时(两位)分(两位).时效(三位)”；缺省-YYYYMMDDHHFF.000.m3)
        show_url:  是否完整显示url，默认为不显示
        defaultNan: 缺测显示值，默认缺测存成9999
        dataFormat: 返回数据格式，默认为json
        url: cmadaas的url前面固定部分，默认'http://10.40.17.54/music-ws/api?'
        SN_id: cmadaas服务节点代码，默认'NMIC_MUSIC_CMADAAS'
        ###
        """
        import os
        ## 根据参数字典，生成url
        LogLib.info('CMADaasAccess OBS output Starts: {0}'.format(datetime.datetime.now().strftime("%Y%m%d%H%M")))
        time = CMADaasAccess.get_datetime_from_str(time)
        #print(time)
        urlstr = CMADaasAccess.combine_url_from_para(qparams,time,**kargs)
        if show:print(urlstr)
        ## 得到url返回结果
        json = CMADaasAccess.read_data_from_cmadaas(urlstr)
        ## 生成meteva数据
        sta = CMADaasAccess.dataframe_from_dict_json(json,default=defaultNan)
        ## 输出micaps3格式
        if save_path is not None:
            try:
                if filename is None: filename="YYYYMMDDHHFF.000.m3"
                output_name = os.path.join(save_path, filename)
                output_name = meteva.base.get_path(output_name,time)
                print('OUTPUT: ', output_name)
                meteva.base.write_stadata_to_micaps3(sta, save_path=output_name, effectiveNum=2)
            except Exception as data:
                LogLib.error('{0} CMADaasAccess OBS output ERROR!: {1}'.format(datetime.datetime.now().strftime("%Y%m%d%H%M"),data))        
        return(sta)


            
    @staticmethod
    def get_files_from_cmadaas(qparams, time='202011110200', save_path=None, filename=None, show_files=True, is_new=True,
                               **kargs):
        """
        time： 可以为字符串str，如“202011110200”；也可以为datetime类时间数据 
        save_path: 存储文件目录，若不填写则不存储
        filename:  存储文件名,可以使用通配符(如YYYYMMDDHHFF.TTT即表示“年(四位)月(两位)日(两位)时(两位)分(两位).时效(三位)”；缺省-使用云平台原始文件名)
        show_url: 是否完整显示url，默认为不显示
        show_files: 是否完整显示待下载文件，默认为不显示
        is_new    ；True-只处理下载最新入库时间的一个文件;Fales-处理所有文件
        """
        import os
        import pandas as pd
        import datetime
        ## 根据参数字典，生成url
        time = CMADaasAccess.get_datetime_from_str(time)
        urlstr = CMADaasAccess.combine_url_from_para(qparams,time,**kargs)
        ## 得到url返回结果
        json = CMADaasAccess.read_data_from_cmadaas(urlstr)
        df = pd.DataFrame(json)
        try:
            df.sort_values(by='IYMDHM',inplace=True,ascending=False)#按入库时间降序排列，新更新的在上方
        except:
            print("no key IYMDHM")
        df.reset_index(drop=True,inplace=True)
        if is_new : df = df.loc[0].to_frame().T
        if show_files : print(df,type(df))
        
        try:
            LogLib.info('FixWriteData write_data_to_file start %s' % (save_path))
            if save_path is None: return(json)#save_path为None，不存储文件，直接返回json信息
            
            for index in range(len(df)):
                fname = df.loc[index,'FILE_NAME']
                url = df.loc[index,'FILE_URL']
                datetime = df.loc[index, 'Datetime']
                ## url内容读取并文件保存
                file_time = CMADaasAccess.get_datetime_from_str(datetime)
                # 原url文件名
                if filename is None:
                    output_name = os.path.join(save_path, fname)
                # 自定义文件名格式
                else:
                    output_name = os.path.join(save_path, filename)
                output_name = meteva.base.get_path(output_name,file_time)
                # 文件存储
                grd,header = CMADaasAccess.read_file_from_cmadaas(url)
                print('WriteFile : ',output_name)
                CMADaasAccess.write_data_to_file(output_name,grd,creat_dir=True)
                
        except Exception as data:
            print("ERROR:",data)
            LogLib.error('{0} CMADaasAccess Files output ERROR!: {1}'.format(datetime.datetime.now().strftime("%Y%m%d%H%M"),data))
        
        return(json)
#        
        
            
            
        

if __name__ == '__main__':

    dati_list = [datetime.datetime(2021,12,1,0,0),datetime.datetime(2021,12,1,1,0)]
    qparams = {'serviceNodeId':'NMIC_MUSIC_CMADAAS',
            'userId':'NMC_SGFV',
            'interfaceId':'getSateFileByTimeRange',
            'dataCode':'SATE_FY4A_STA_L1',
            'timeRange':CMADaasAccess.get_timerange(dati_list),
            'pwd':'NMC_yfsjyk_601'
            }
    save_dir = r"H:\test_data\output\meb\get_files"
    urlstr = CMADaasAccess.get_files_from_cmadaas(qparams,show_url=True,save_path=save_dir)
    print(urlstr)


